# Copyright 2017 fmnisme@gmail.com christian@jonak.org
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# @author: Christian Jonak-Moechel, fmnisme, Tobias von der Krone
# @contact: christian@jonak.org, fmnisme@gmail.com, tobias@vonderkrone.info
# @summary: Python library for the Icinga 2 RESTful API

"""
Provides the base class :class:`RequestHandler` that is inherited by the different
endpoint classes, for example the class ``Events`` handles the ``v1/events`` endpoint,
the class ``Objects`` handles the ``v1/objects`` entpoint and so on...
"""

import json
from typing import (
    TYPE_CHECKING,
    Any,
    Generator,
    Optional,
)
from urllib.parse import urljoin

import requests
import urllib3

from pretiac.config import Config
from pretiac.exceptions import PretiacException, PretiacRequestException
from pretiac.object_types import (
    HostState,
    Payload,
    RequestMethod,
    ServiceState,
    State,
)

if TYPE_CHECKING:
    from pretiac.raw_client import RawClient

# https://stackoverflow.com/a/28002687
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def normalize_state(state: State | Any) -> int:
    if isinstance(state, ServiceState) or isinstance(state, HostState):
        return state.value
    if isinstance(state, int) and 0 <= state <= 3:
        return state
    raise PretiacException("invalid")


class RequestHandler:
    """
    Handles the HTTP requests to the Icinga2 API.
    """

    raw_client: "RawClient"

    api_version: str = "v1"

    path_prefix: Optional[str] = None
    """
    for example ``objects`` for ``localhost:5665/v1/objects``
    """

    def __init__(self, client: "RawClient") -> None:
        """
        initialize object
        """

        self.raw_client = client
        self.stream_cache = ""

    @property
    def versioned_path_prefix(self) -> str:
        if not self.path_prefix:
            raise PretiacException("Specify self.path_prefix")
        return f"{self.api_version}/{self.path_prefix}"

    @property
    def config(self) -> Config:
        return self.raw_client.config

    def __create_session(self, method: RequestMethod = "POST") -> requests.Session:
        """
        Create a session object.
        """

        session = requests.Session()
        # prefer certificate authentification
        if self.config.client_certificate and self.config.client_private_key:
            # The certificate and RSA private key are in different files.
            session.cert = (
                self.config.client_certificate,
                self.config.client_private_key,
            )
        elif self.config.client_certificate:
            # The certificate and RSA private key are in the same file.
            session.cert = self.config.client_certificate
        elif self.config.http_basic_username and self.config.http_basic_password:
            # use username and password
            session.auth = (
                self.config.http_basic_username,
                self.config.http_basic_password,
            )
        session.headers = {
            "User-Agent": f"Python-pretiac/{self.raw_client.version}",
            "X-HTTP-Method-Override": method.upper(),
            "Accept": "application/json",
        }

        return session

    def __throw_exception(self, suppress_exception: Optional[bool] = None) -> bool:
        if isinstance(suppress_exception, bool):
            return not suppress_exception
        if isinstance(self.config.suppress_exception, bool):
            return not self.config.suppress_exception
        return True

    def _request(
        self,
        method: RequestMethod,
        url_relpath: Optional[str],
        payload: Optional[dict[str, Any]] = None,
        stream: bool = False,
        suppress_exception: Optional[bool] = None,
    ) -> Any:
        """
        make the request and return the body

        :param method: The HTTP method, for example ``GET``, ``POST``
        :param url_relpath: The requested url path without the path prefix. If
            you want to query the URL
            ``https://localhost:5665/v1/objects/hosts`` specify only ``hosts``.
        :param payload: The payload to send
        :param suppress_exception: If this parameter is set to ``True``, no exceptions are thrown.

        :returns: The response as json
        """

        request_url = urljoin(
            self.raw_client.url,
            self.versioned_path_prefix
            if url_relpath is None
            else f"{self.versioned_path_prefix}/{url_relpath}",
        )

        # create session
        session = self.__create_session(method)

        # create arguments for the request
        request_args: Payload = {"url": request_url}
        if payload:
            request_args["json"] = payload
        if self.config.ca_certificate:
            request_args["verify"] = self.config.ca_certificate
        else:
            request_args["verify"] = False
        if stream:
            request_args["stream"] = True

        # do the request
        response: requests.Response = session.post(**request_args)

        if not stream:
            session.close()

        if (
            self.__throw_exception(suppress_exception)
            and not 200 <= response.status_code <= 299
        ):
            raise PretiacRequestException(
                f'Request "{response.url}" failed with status {response.status_code}: {response.text}',
                response.json(),
            )

        if stream:
            return response
        else:
            return response.json()

    @staticmethod
    def _get_message_from_stream(
        stream: requests.Response,
    ) -> Generator[Any, Any, None]:
        """
        Make the request and return the body.

        :param stream: The stream.

        :returns: The message.
        """
        for line in stream.iter_lines():
            yield json.loads(line)



# Copyright 2017 fmnisme@gmail.com christian@jonak.org
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# @author: Christian Jonak-Moechel, fmnisme, Tobias von der Krone
# @contact: christian@jonak.org, fmnisme@gmail.com, tobias@vonderkrone.info
# @summary: Python library for the Icinga 2 RESTful API

"""
Icinga 2 API client exceptions
"""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pretiac.request_handler import Payload


class PretiacException(Exception):
    """
    Icinga 2 API exception class
    """


class PretiacRequestException(PretiacException):
    """
    Icinga 2 API Request exception class
    """

    response: "Payload" = {}

    def __init__(self, message: str, response: "Payload") -> None:
        super(PretiacRequestException, self).__init__(message)
        self.response = response


class PretiacConfigFileException(Exception):
    """
    Icinga 2 API config file exception class
    """



# Copyright 2017 fmnisme@gmail.com christian@jonak.org
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# @author: Christian Jonak-Moechel, fmnisme, Tobias von der Krone
# @contact: christian@jonak.org, fmnisme@gmail.com, tobias@vonderkrone.info
# @summary: Python library for the Icinga 2 RESTful API

"""
pretiac is a `Python <http://www.python.org>`_ module to interact with the
`Icinga 2 RESTful API <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/>`_.
"""

from pathlib import Path
from typing import Optional

from pretiac.client import Client

__client: Optional[Client] = None


def set_default_client(
    config_file: Optional[str | Path] = None,
    api_endpoint_host: Optional[str] = None,
    api_endpoint_port: Optional[int] = None,
    http_basic_username: Optional[str] = None,
    http_basic_password: Optional[str] = None,
    client_private_key: Optional[str] = None,
    client_certificate: Optional[str] = None,
    ca_certificate: Optional[str] = None,
    suppress_exception: Optional[bool] = None,
) -> Client:
    """
    Set and configure the default client.

    :param config_file: The path of the configuration file to load.
    :param api_endpoint_host: The domain or the IP address of the API
        endpoint, e. g. ``icinga.example.com``, ``localhost`` or ``127.0.0.1``.
    :param api_endpoint_port: The TCP port of the API endpoint, for example
        ``5665``.
    :param http_basic_username: The name of the API user used in the HTTP basic
        authentification, e. g. ``apiuser``.
    :param http_basic_password: The password of the API user used in the HTTP
        basic authentification, e. g. ``password``.
    :param client_private_key: The file path of the client’s **private RSA
        key**, for example ``/etc/pretiac/api-client.key.pem``.
    :param client_certificate: The file path of the client’s **certificate**,
        for example ``/etc/pretiac/api-client.cert.pem``.
    :param ca_certificate: The file path of the Icinga **CA (Certification
        Authority)**, for example ``/var/lib/icinga2/certs/ca.crt``.
    :param suppress_exception: If set to ``True``, no exceptions are thrown.
    """
    global __client
    __client = Client(
        config_file=config_file,
        api_endpoint_host=api_endpoint_host,
        api_endpoint_port=api_endpoint_port,
        http_basic_username=http_basic_username,
        http_basic_password=http_basic_password,
        client_private_key=client_private_key,
        client_certificate=client_certificate,
        ca_certificate=ca_certificate,
        suppress_exception=suppress_exception,
    )
    return __client


def get_default_client() -> Client:
    """
    Get the default client.

    This function intentionally has no input parameters. Use the
    function :func:`set_default_client` to set a new configured client.

    However, this function loads a client configured by configuration files in
    the following order (if the function :func:`set_default_client` was not called before):

    1. The file path in the environment variable ``PRETIAC_CONFIG_FILE``.
    2. The configuration file in the home folder ``~/.pretiac.yml``.
    3. The configuration file at ``/etc/pretiac/config.yml``.

    .. code-block:: yaml

        ---
        api_endpoint_host: localhost
        api_endpoint_port: 5665
        client_private_key: /etc/pretiac/api-client.key.pem
        client_certificate: /etc/pretiac/api-client.cert.pem
        ca_certificate: /etc/pretiac/ca.crt
        new_host_defaults:
            templates: [passive-host]
        new_service_defaults:
            templates: [passive-service]
            attrs:
                check_interval: monthly

    """
    global __client
    if not __client:
        __client = Client()
    return __client



from logging import (
    DEBUG,
    INFO,
    Formatter,
    StreamHandler,
    basicConfig,
    getLogger,
)
from logging import (
    Logger as DefaultLogger,
)


class Logger:
    """A wrapper around the Python logging module with 3 debug logging levels.

    1. ``-d``: info
    2. ``-dd``: debug
    3. ``-ddd``: verbose

    :see `check_systemd.py L289-L351 <https://github.com/Josef-Friedrich/check_systemd/blob/a63450608240e3d0432a2a6ad7501f79e5e0dce9/check_systemd.py#L289-L351>`__:
    """

    __logger: DefaultLogger

    __BLUE = "\x1b[0;34m"
    __PURPLE = "\x1b[0;35m"
    __CYAN = "\x1b[0;36m"
    __RESET = "\x1b[0m"

    __INFO: int = INFO
    __DEBUG: int = DEBUG
    __VERBOSE = 5

    def __init__(self) -> None:
        handler = StreamHandler()
        handler.setFormatter(Formatter("%(message)s"))
        basicConfig(handlers=[handler])
        self.__logger = getLogger(__name__)

    def set_level(self, level: int) -> None:
        # NOTSET=0
        # custom level: VERBOSE=5
        # DEBUG=10
        # INFO=20
        # WARN=30
        # ERROR=40
        # CRITICAL=50
        if level == 1:
            self.__logger.setLevel(INFO)
        elif level == 2:
            self.__logger.setLevel(DEBUG)
        elif level > 2:
            self.__logger.setLevel(5)

    def __log(self, level: int, color: str, msg: str, *args: object) -> None:
        a: list[str] = []
        for arg in args:
            a.append(color + str(arg) + self.__RESET)
        self.__logger.log(level, msg, *a)

    def info(self, msg: str, *args: object) -> None:
        """Log on debug level ``1``: ``-d``.

        :param msg: A message format string. Note that this means that you can
            use keywords in the format string, together with a single
            dictionary argument. No ``%`` formatting operation is performed on
            ``msg`` when no args are supplied.
        :param args: The arguments which are merged into ``msg`` using the
            string formatting operator.
        """
        self.__log(self.__INFO, self.__BLUE, msg, *args)

    def debug(self, msg: str, *args: object) -> None:
        """Log on debug level ``2``: ``-dd``.

        :param msg: A message format string. Note that this means that you can
            use keywords in the format string, together with a single
            dictionary argument. No ``%`` formatting operation is performed on
            ``msg`` when no args are supplied.
        :param args: The arguments which are merged into ``msg`` using the
            string formatting operator.
        """
        self.__log(self.__DEBUG, self.__PURPLE, msg, *args)

    def verbose(self, msg: str, *args: object) -> None:
        """Log on debug level ``3``: ``-ddd``

        :param msg: A message format string. Note that this means that you can
            use keywords in the format string, together with a single
            dictionary argument. No ``%`` formatting operation is performed on
            ``msg`` when no args are supplied.
        :param args: The arguments which are merged into ``msg`` using the
            string formatting operator.
        """
        self.__log(self.__VERBOSE, self.__CYAN, msg, *args)

    def show_levels(self) -> None:
        msg = "log level %s (%s): %s"
        self.info(msg, 1, "info", "-d")
        self.debug(msg, 2, "debug", "-dd")
        self.verbose(msg, 3, "verbose", "-ddd")


logger = Logger()



from argparse import ArgumentParser

from rich import print

from pretiac import set_default_client
from pretiac.check_executor import check
from pretiac.config import load_config_file
from pretiac.log import logger
from pretiac.object_types import (
    normalize_to_plural_snake_object_type_name,
)


def main() -> None:
    client = set_default_client()
    parser = ArgumentParser(
        prog="icinga-api",
        description="Command line interface for the Icinga2 API.",
    )

    # global options
    parser.add_argument(
        "-d",
        "--debug",
        action="count",
        default=0,
        help="Increase debug verbosity (use up to 3 times): -d: info -dd: debug -ddd: verbose.",
    )

    sub_parsers = parser.add_subparsers(dest="sub_command", help="sub-command help")

    # check
    check_parser = sub_parsers.add_parser(
        "check", help="Execute checks and send it to the monitoring server."
    )

    check_parser.add_argument("--file")

    # config
    sub_parsers.add_parser("config", help="Dump the configuration")

    # events
    objects_parser = sub_parsers.add_parser(
        "events", help="Subscribe to an event stream."
    )

    # objects
    objects_parser = sub_parsers.add_parser(
        "objects", help="List the different configuration object types."
    )

    objects_parser.add_argument("object_type")

    # send-service-check-result
    send_parser = sub_parsers.add_parser(
        "send-service-check-result",
        help="Send service check results to the specified API endpoint.",
    )

    send_parser.add_argument("service")

    send_parser.add_argument("--host")

    send_parser.add_argument("--exit-status")

    send_parser.add_argument("--plugin-output")

    send_parser.add_argument("--performance-data")

    # status
    sub_parsers.add_parser(
        "status",
        help="Retrieve status information and statistics for Icinga 2.",
    )

    args = parser.parse_args()

    logger.set_level(args.debug)
    logger.show_levels()

    if args.sub_command == "check":
        check(args.file)

    elif args.sub_command == "config":
        config = load_config_file()
        print(config)

    elif args.sub_command == "events":
        for event in client.subscribe_events(["CheckResult"], "cli"):
            print(event)

    elif args.sub_command == "objects":
        print(
            getattr(
                client,
                f"get_{normalize_to_plural_snake_object_type_name(args.object_type)}",
            )()
        )

    elif args.sub_command == "send-service-check-result":
        print(
            client.send_service_check_result(
                service=args.service,
                host=args.host,
                exit_status=args.exit_status,
                plugin_output=args.plugin_output,
                performance_data=args.performance_data,
            )
        )

    elif args.sub_command == "status":
        print(client.get_status())



"""
All available config object types.

Listed in the same order as in this `Markdown document <https://github.com/Icinga/icinga2/blob/master/doc/09-object-types.md>`__.
"""

import re
from collections.abc import Sequence
from enum import Enum
from typing import (
    Annotated,
    Any,
    Literal,
    Optional,
    TypeAlias,
    Union,
    get_args,
)

from pydantic import BeforeValidator
from pydantic.dataclasses import dataclass


def _empty_str_to_none(v: str | None) -> str | None:
    if v is None:
        return None
    if v == "":
        return None
    return v


# https://github.com/pydantic/pydantic/discussions/2687#discussioncomment-9893991
OptionalStr: TypeAlias = Annotated[Optional[str], BeforeValidator(_empty_str_to_none)]
"""
An empty string is set to None by the Pydantic validator.
"""


MonitoringObjectName = Literal[
    "ApiUser",
    "CheckCommand",
    # "CheckCommandArguments", used in CheckCommand
    "Dependency",
    "Endpoint",
    "EventCommand",
    "Host",
    "HostGroup",
    "Notification",
    "NotificationCommand",
    "ScheduledDowntime",
    "Service",
    "ServiceGroup",
    "TimePeriod",
    "User",
    "UserGroup",
    "Zone",
]
"""
see `doc/09-object-types.md object-types-monitoring <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md#object-types-monitoring>`__
"""


RuntimeObjectName = Literal["Comment", "Downtime"]
"""
see [doc/09-object-types.md runtime-objects-](https://github.com/Icinga/icinga2/b
"""

FeatureObjectName = Literal[
    "ApiListener",
    "CheckerComponent",
    "CompatLogger",
    "ElasticsearchWriter",
    "ExternalCommandListener",
    "FileLogger",
    "GelfWriter",
    "GraphiteWriter",
    "IcingaApplication",
    "IcingaDB",
    "IdoMySqlConnection",
    "IdoPgsqlConnection",
    "InfluxdbWriter",
    "Influxdb2Writer",
    "JournaldLogger",
    "LiveStatusListener",
    "NotificationComponent",
    "OpenTsdbWriter",
    "PerfdataWriter",
    "SyslogLogger",
    "WindowsEventLogLogger",
]
"""
see `doc/09-object-types.md features- <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md#features->`__
"""


ObjectTypeName = Union[MonitoringObjectName, RuntimeObjectName, FeatureObjectName]


def _convert_pascal_to_snake_case(name: str) -> str:
    """
    Insert underscores before capital letters and convert to lowercase
    """
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def get_object_types_names() -> list[str]:
    return (
        list(get_args(MonitoringObjectName))
        + list(get_args(RuntimeObjectName))
        + list(get_args(FeatureObjectName))
    )


object_type_names: list[str] = get_object_types_names()

object_type_names_snake = list(
    map(lambda name: _convert_pascal_to_snake_case(name), object_type_names)
)


def normalize_to_plural_snake_object_type_name(object_type_name: str) -> str:
    """
    for example: ``ApiUser`` to ``api_users``

    :returns: A pluralized object type name in the snake case.
    """
    snake = _convert_pascal_to_snake_case(object_type_name)
    if snake in object_type_names_snake:
        return f"{snake}s"
    return snake


def pluralize_to_lower_object_type_name(object_type_name: ObjectTypeName) -> str:
    """
    for example: ``ApiUser`` to ``apiusers``
    """
    return f"{object_type_name.lower()}s"


HostOrService = Literal["Host", "Service"]

HostServiceComment = Union[Literal["Comment"], HostOrService]

HostServiceDowntime = Union[Literal["Downtime"], HostOrService]


Payload = dict[str, Any]

FilterVars = Optional[Payload]

RequestMethod = Literal["GET", "OPTIONS", "HEAD", "POST", "PUT", "PATCH", "DELETE"]
"""
https://github.com/psf/requests/blob/a3ce6f007597f14029e6b6f54676c34196aa050e/src/requests/api.py#L17

https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods
"""


########################################################################################
# Delegated interfaces and types
########################################################################################


@dataclass
class SourceLocation:
    """https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/remote/templatequeryhandler.cpp#L29-L35"""

    path: str
    """
    ``/etc/icinga2-custom/conf.d/api-users.conf``
    """

    first_line: int

    first_column: int

    last_line: int

    last_column: int


class HAMode(Enum):
    """
    ``HA`` = High-Availability

    :see: `lib/base/configobject.ti L12-L16 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/base/configobject.ti#L12-L16>`__"""

    HARunOnce = 0
    HARunEverywhere = 1


class StateType(Enum):
    """
    see: `lib/icinga/checkresult.ti L38-L43 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/checkresult.ti#L38-L43>`__
    """

    StateTypeSoft = 0
    StateTypeHard = 1


@dataclass
class Dictionary:
    pass


Timestamp = float
"""
for example `1699475880.364077`

:see: `lib/base/value.hpp L15 <https://github.com/Icinga/icinga2/blob/4c6b93d61775ff98fc671b05ad4de2b62945ba6a/lib/base/value.hpp#L15>`__
"""


Value = int | float | str | bool | Any
"""
A type that can hold an arbitrary value.

`lib/base/value.hpp L31-L145 <https://github.com/Icinga/icinga2/blob/4c6b93d61775ff98fc671b05ad4de2b62945ba6a/lib/base/value.hpp#L31-L145>`_
"""


class ServiceState(Enum):
    """

    0=OK, 1=WARNING, 2=CRITICAL, 3=UNKNOWN

    https://github.com/Icinga/icinga2/blob/a8adfeda60232260e3eee6d68fa5f4787bb6a245/lib/icinga/checkresult.ti#L22-L33
    """

    OK = 0
    WARNING = 1
    CRITICAL = 2
    UNKNOWN = 3

    def __str__(self) -> str:
        return f"{self.value} ({self.name})"


class HostState(Enum):
    """
     0=UP, 1=DOWN.

    https://github.com/Icinga/icinga2/blob/a8adfeda60232260e3eee6d68fa5f4787bb6a245/lib/icinga/checkresult.ti#L11-L20
    """

    UP = 0
    DOWN = 1

    def __str__(self) -> str:
        return f"{self.value} ({self.name})"


State = HostState | ServiceState | Literal[0, 1, 2, 3] | int


def get_service_state(state: Union[State, Any]) -> ServiceState:
    if isinstance(state, ServiceState):
        return state
    if isinstance(state, int) and 0 <= state <= 3:
        return ServiceState(state)
    return ServiceState.CRITICAL


@dataclass
class CheckResult:
    """

    The attributes are as listed in the offical Icinga2 documentation.

    :see: `doc/08-advanced-topics/#checkresult <https://icinga.com/docs/icinga-2/latest/doc/08-advanced-topics/#checkresult>`__
    :see: `lib/icinga/checkresult.ti <https://github.com/Icinga/icinga2/blob/master/lib/icinga/checkresult.ti>`__
    """

    type = "CheckResult"

    exit_status: int
    """
    The exit status returned by the check execution.

    https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/icinga/checkresult.ti#L55
    """

    output: str
    """
    The check output.

    https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/icinga/checkresult.ti#L59
    """

    performance_data: Optional[list[str]]
    """
    Array of performance data values.

    https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/icinga/checkresult.ti#L60
    """

    check_source: str
    """
    Name of the node executing the check.
    """

    scheduling_source: str
    """
    Name of the node scheduling the check.
    """

    state: ServiceState
    """
    The current state (0 = OK, 1 = WARNING, 2 = CRITICAL, 3 = UNKNOWN).

    https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/icinga/checkresult.ti#L57
    """

    previous_hard_state: Union[ServiceState, int]
    """
    Sometimes ``99``?

    https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/icinga/checkresult.ti#L58C1-L58C49
    """

    command: Optional[Union[list[str], str]]
    """
    Array of command with shell-escaped arguments or command line string.

    https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/icinga/checkresult.ti#L54
    """

    execution_start: Timestamp
    """
    Check execution start time (as a UNIX timestamp).

    https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/icinga/checkresult.ti#L51
    """

    execution_end: Timestamp
    """
    Check execution end time (as a UNIX timestamp).

    https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/icinga/checkresult.ti#L52
    """

    schedule_start: Timestamp
    """
    Scheduled check execution start time (as a UNIX timestamp).

    https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/icinga/checkresult.ti#L49
    """

    schedule_end: Timestamp
    """
    Scheduled check execution end time (as a UNIX timestamp).

    https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/icinga/checkresult.ti#L50
    """

    active: bool
    """
    Whether the result is from an active or passive check.

    https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/icinga/checkresult.ti#L62-L64
    """

    vars_before: Optional[dict[str, Any]]
    """
    Internal attribute used for calculations.

    https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/icinga/checkresult.ti#L70
    """

    vars_after: dict[str, Any]
    """
 	Internal attribute used for calculations.

    https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/icinga/checkresult.ti#L71
    """

    ttl: int
    """
    Time-to-live duration in seconds for this check result. The next expected
    check result is ``now + ttl`` where freshness checks are executed.

    https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/icinga/checkresult.ti#L68
    """


########################################################################################
# Interface from which the object types inherit
########################################################################################


@dataclass(config={"extra": "forbid"})
class ConfigObject:
    """
    :see: `lib/base/configobject.ti L57-L92 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/base/configobject.ti#L57-L92>`__
    """

    name: Optional[str] = None
    """:see: `lib/base/configobject.ti L59-L68 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/base/configobject.ti#L59-L68>`__"""

    type: Optional[str] = None

    zone: OptionalStr = None
    """:see: `lib/base/configobject.ti L69 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/base/configobject.ti#L69>`__"""

    package: Optional[str] = None
    """for example ``_etc``

    :see: `lib/base/configobject.ti L70 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/base/configobject.ti#L70>`__"""

    templates: Optional[Sequence[str]] = None
    """:see: `lib/base/configobject.ti L71 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/base/configobject.ti#L71>`__"""

    source_location: Optional[SourceLocation] = None
    """
    :see: `lib/base/configobject.ti L72-L74 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/base/configobject.ti#L72-L74>`__
    """

    active: Optional[bool] = None
    """:see: `lib/base/configobject.ti L75 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/base/configobject.ti#L75>`__"""

    paused: Optional[bool] = None
    """:see: `lib/base/configobject.ti L76-L78 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/base/configobject.ti#L76-L78>`__"""

    ha_mode: Optional[HAMode] = None
    """:see: `lib/base/configobject.ti L83 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/base/configobject.ti#L83>`__"""

    original_attributes: Optional[dict[str, Any]] = None
    """:see: `lib/base/configobject.ti L87 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/base/configobject.ti#L87>`__"""

    version: Optional[float] = None
    """:see: `lib/base/configobject.ti L88-L90 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/base/configobject.ti#L88-L90>`__"""


@dataclass(config={"extra": "forbid"})
class CustomVarObject(ConfigObject):
    """
    :see: `lib/icinga/customvarobject.ti L10 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/customvarobject.ti#L10>`__
    """

    vars: Optional[dict[str, Any]] = None
    """
    :see: `lib/icinga/customvarobject.ti L12 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/customvarobject.ti#L12>`__
    """


@dataclass(config={"extra": "forbid"})
class Checkable(CustomVarObject):
    """
    :see: `lib/icinga/checkable.ti <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/checkable.ti>`__
    """

    check_command: Optional[str] = None
    """
    The name of the check command.

    :see: `doc/09-object-types.md L717 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L717>`__
    """

    max_check_attempts: Optional[int] = None
    """
    The float of times a service is re-checked before changing into a hard state. Defaults to 3.

    :see: `doc/09-object-types.md L718 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L718>`__
    """

    check_period: Optional[str] = None
    """
    The name of a time period which determines when this service should be checked. Not set by default (effectively 24x7).

    :see: `doc/09-object-types.md L719 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L719>`__

    """

    check_timeout: Optional[Value] = None
    """
    Check command timeout in seconds. Overrides the CheckCommand's `timeout` attribute.

    :see: `doc/09-object-types.md L720 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L720>`__
    """

    check_interval: Optional[float] = None
    """
    The check interval (in seconds). This interval is used for checks when the service is in a `HARD` state. Defaults to `5m`.

    :see: `doc/09-object-types.md L721 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L721>`__
    """

    retry_interval: Optional[float] = None
    """
    This interval is used for checks when the service is in a `SOFT` state. Defaults to `1m`. Note: This does not affect the scheduling `after a passive check result <08-advanced-topics.md#check-result-freshness>`__.
    """

    event_command: Optional[str] = None

    volatile: Optional[bool] = None

    enable_active_checks: Optional[bool] = None

    enable_passive_checks: Optional[bool] = None

    enable_event_handler: Optional[bool] = None

    enable_notifications: Optional[bool] = None

    enable_flapping: Optional[bool] = None

    enable_perfdata: Optional[bool] = None

    flapping_ignore_states: Optional[Sequence[str]] = None

    flapping_threshold: Optional[float] = None
    """
    deprecated
    """
    flapping_threshold_low: Optional[float] = None

    flapping_threshold_high: Optional[float] = None

    notes: OptionalStr = None
    """ Optional. Notes for the checkable."""

    notes_url: OptionalStr = None
    """Optional. URL for notes for the checkable (for example, in notification commands)."""

    action_url: OptionalStr = None
    """Optional. URL for actions for the checkable (for example, an external graphing tool)."""

    icon_image: OptionalStr = None
    """Optional. Icon image for the checkable. Used by external interfaces only."""

    icon_image_alt: OptionalStr = None
    """Optional. Icon image description for the checkable. Used by external interface only."""

    next_check: Optional[Timestamp] = None

    check_attempt: Optional[int] = None

    state_type: Optional[StateType] = None

    last_state_type: Optional[StateType] = None

    last_reachable: Optional[bool] = None

    last_check_result: Optional[CheckResult] = None

    last_state_change: Optional[Timestamp] = None

    last_hard_state_change: Optional[Timestamp] = None

    last_state_unreachable: Optional[Timestamp] = None

    previous_state_change: Optional[Timestamp] = None

    severity: Optional[int] = None

    problem: Optional[bool] = None

    handled: Optional[bool] = None

    next_update: Optional[Timestamp] = None

    force_next_check: Optional[bool] = None

    acknowledgement: Optional[int] = None

    acknowledgement_expiry: Optional[Timestamp] = None

    acknowledgement_last_change: Optional[Timestamp] = None

    force_next_notification: Optional[bool] = None

    last_check: Optional[Timestamp] = None

    downtime_depth: Optional[int] = None

    flapping_current: Optional[float] = None

    flapping_last_change: Optional[Timestamp] = None

    flapping: Optional[bool] = None

    command_endpoint: Optional[str] = None

    executions: Optional[Dictionary] = None


########################################################################################
# The individual object types
########################################################################################

########################################################################################
# Monitoring Objects
########################################################################################


@dataclass(config={"extra": "forbid"})
class ApiUser(ConfigObject):
    """
    ApiUser objects are used for authentication against the `Icinga 2 API <12-icinga2-api.md#icinga2-api-authentication>`__.


    .. code-block::

        object ApiUser "root" {
            password = "mysecretapipassword"
            permissions = [ "*" ]
        }

    .. tags:: Object type, Monitoring object type

    https://icinga.com/docs/icinga-2/latest/doc/09-object-types/#apiuser
    :see: `lib/remote/apiuser.ti <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/remote/apiuser.ti>`__
    :see: `doc/09-object-types.md L41-L63 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L41-L63>`__
    """

    password: Optional[str] = None
    """
    Password string. Note: This attribute is hidden in API responses.

    :see: `lib/remote/apiuser.ti L14 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/remote/apiuser.ti#L14>`__
    """

    client_cn: Optional[str] = None
    """
    Client Common Name (CN).

    .. tags:: config

    :see: `lib/remote/apiuser.ti L16 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/remote/apiuser.ti#L16>`__
    """

    permissions: Optional[Sequence[str]] = None
    """
    Array of permissions. Either as string or dictionary with the keys `permission` and `filter`. The latter must be specified as function.

    .. tags:: config

    :see: `lib/remote/apiuser.ti L17 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/remote/apiuser.ti#L17>`__
    :see: `lib/remote/apiuser.ti L21-L28 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/remote/apiuser.ti#L21-L28>`__
    """


@dataclass(config={"extra": "forbid"})
class Function:
    """
    https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/base/function.ti#L10-L16
    """

    type: str
    name: str
    side_effect_free: bool
    deprecated: bool
    arguments: Sequence[str]


@dataclass(config={"extra": "forbid"})
class CheckCommandArguments:
    """
    Command arguments can be defined as key-value-pairs in the `arguments`
    dictionary. Best practice is to assign a dictionary as value which
    provides additional details such as the `description` next to the `value`.

    Example

    .. code-block::

        arguments = {
            "--parameter" = {
                description = "..."
                value = "..."
            }
        }

    .. tags:: Object type, Monitoring object type

    :see: `doc/09-object-types.md L117-L150 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L117-L150>`__
    :see: `lib/icinga/command.ti L30-L46 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/command.ti#L30-L46>`__
    :see: `lib/icinga/command.ti L33-L45 <https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/icinga/command.ti#L33-L45>`__
    """

    key: Optional[str] = None
    value: Optional[str | Function | Any] = None
    description: Optional[str] = None
    required: Optional[bool] = None
    skip_key: Optional[bool] = None
    repeat_key: Optional[bool] = None
    set_if: Optional[str | Function] = None
    order: Optional[float] = None
    separator: Optional[str] = None


@dataclass(config={"extra": "forbid"})
class CheckCommand(CustomVarObject):
    """
    A check command definition. Additional default command custom variables can be
    defined here.

    Example

    .. code-block::

        object CheckCommand "http" {
            command = [ PluginDir + "/check_http" ]

            arguments = {
                "-H" = "$http_vhost$"
                "-I" = "$http_address$"
                "-u" = "$http_uri$"
                "-p" = "$http_port$"
                "-S" = {
                    set_if = "$http_ssl$"
                }
                "--sni" = {
                    set_if = "$http_sni$"
                }
                "-a" = {
                    value = "$http_auth_pair$"
                    description = "Username:password on sites with basic authentication"
                }
                "--no-body" = {
                    set_if = "$http_ignore_body$"
                }
                "-r" = "$http_expect_body_regex$"
                "-w" = "$http_warn_time$"
                "-c" = "$http_critical_time$"
                "-e" = "$http_expect$"
            }

            vars.http_address = "$address$"
            vars.http_ssl = false
            vars.http_sni = false
        }

    https://icinga.com/docs/icinga-2/latest/doc/09-object-types/#checkcommand
    :see: `doc/09-object-types.md L65-L114 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L65-L114>`__
    :see: `lib/icinga/command.ti <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/command.ti>`__

    .. tags:: Object type, Monitoring object type
    """

    timeout: Optional[int] = None
    """
    **Optional.** The command timeout in seconds.  Defaults to `1m`.

    https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/command.ti#L15-L17
    """

    execute: Optional[Union[str, Function]] = None
    """
    https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/command.ti#L19
    """

    command: Optional[Union[Sequence[str], Payload]] = None
    """
    **Required.** The command. This can either be an array of individual
    command arguments. Alternatively a string can be specified in which case
    the shell interpreter (usually /bin/sh) takes care of parsing the command.
    When using the "arguments" attribute this must be an array. Can be specified
    as function for advanced implementations.

    https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/command.ti#L25-L28
    """

    arguments: Optional[
        Union[dict[str, Union[CheckCommandArguments, str, Sequence[str], Payload]], str]
    ] = None
    """
    **Optional.** A dictionary of command arguments.

    https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/command.ti#L30-L46
    """

    env: Optional[Any] = None
    """
    **Optional.** A dictionary of macros which should be exported as environment variables prior to executing the command.

    https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/command.ti#L48-L51
    """

    vars: dict[str, Any] | None = None
    """**Optional.** A dictionary containing custom variables that are specific to this command."""


@dataclass(config={"extra": "forbid"})
class Dependency(CustomVarObject):
    """
    Dependency objects are used to specify dependencies between hosts and services. Dependencies
    can be defined as Host-to-Host, Service-to-Service, Service-to-Host, or Host-to-Service
    relations.

    Service-to-Service Example:

    .. code-block::

        object Dependency "webserver-internet" {
            parent_host_name = "internet"
            parent_service_name = "ping4"

            child_host_name = "webserver"
            child_service_name = "ping4"

            states = [ OK, Warning ]

            disable_checks = true
        }

    Host-to-Host Example:

    .. code-block::

        object Dependency "webserver-internet" {
            parent_host_name = "internet"

            child_host_name = "webserver"

            states = [ Up ]

            disable_checks = true
        }

    https://icinga.com/docs/icinga-2/latest/doc/09-object-types/#dependency
    :see: `doc/09-object-types.md L153-L258 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L153-L258>`__
    https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/icinga/dependency.ti#L21-L99

    .. tags:: Object type, Monitoring object type
    """

    parent_host_name: Optional[str] = None
    """Required. The parent host."""

    parent_service_name: OptionalStr = None
    """Optional. The parent service. If omitted, this dependency object is
    treated as host dependency."""

    child_host_name: Optional[str] = None
    """Required. The child host."""

    child_service_name: OptionalStr = None
    """Optional. The child service. If omitted, this dependency object is
    treated as host dependency."""

    redundancy_group: OptionalStr = None
    """Optional. Puts the dependency into a group of mutually redundant
    ones."""

    disable_checks: Optional[bool] = None
    """Optional. Whether to disable checks (i.e., don’t schedule active checks
    and drop passive results) when this dependency fails. Defaults to false."""

    disable_notifications: Optional[bool] = None
    """Optional. Whether to disable notifications when this dependency fails.
    Defaults to true."""

    ignore_soft_states: Optional[bool] = None
    """Optional. Whether to ignore soft states for the reachability calculation.
    Defaults to true."""

    period: OptionalStr = None
    """Optional. Time period object during which this dependency is enabled."""

    states: Optional[Sequence[str]] = None
    """Optional. A list of state filters when this dependency should be OK.
    Defaults to [ OK, Warning ] for services and [ Up ] for hosts."""


@dataclass(config={"extra": "forbid"})
class Endpoint(ConfigObject):
    """
    Endpoint objects are used to specify connection information for remote
    Icinga 2 instances.

    Example

    .. code-block::

        object Endpoint "icinga2-agent1.localdomain" {
            host = "192.168.56.111"
            port = 5665
            log_duration = 1d
        }

    Example (disable replay log):

    .. code-block::

        object Endpoint "icinga2-agent1.localdomain" {
            host = "192.168.5.111"
            port = 5665
            log_duration = 0
        }

    https://icinga.com/docs/icinga-2/latest/doc/09-object-types/#endpoint
    :see: `doc/09-object-types.md L260-L293 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L260-L293>`__
    https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/remote/endpoint.ti#L11-L57

    .. tags:: Object type, Monitoring object type
    """

    host: OptionalStr = None
    """The hostname/IP address of the remote Icinga 2 instance."""

    port: Optional[int] = None
    """The service name/port of the remote Icinga 2 instance. Defaults to 5665."""

    log_duration: Optional[Union[str, int]] = None
    """Optional. Duration for keeping replay logs on connection loss. Defaults to
    1d (86400 seconds). Attribute is specified in seconds. If log_duration is
    set to 0, replaying logs is disabled. You could also specify the value in
    human readable format like 10m for 10 minutes or 1h for one hour."""

    local_log_position: Optional[Timestamp] = None

    remote_log_position: Optional[Timestamp] = None

    icinga_version: Optional[int] = None

    capabilities: Optional[int] = None

    connecting: Optional[bool] = None

    syncing: Optional[bool] = None

    connected: Optional[bool] = None

    last_message_sent: Optional[Timestamp] = None

    last_message_received: Optional[Timestamp] = None

    messages_sent_per_second: Optional[float] = None

    messages_received_per_second: Optional[float] = None

    bytes_sent_per_second: Optional[float] = None

    bytes_received_per_second: Optional[float] = None


@dataclass(config={"extra": "forbid"})
class EventCommand:
    """
    An event command definition.

    .. code-block::

        object EventCommand "restart-httpd-event" {
            command = "/opt/bin/restart-httpd.sh"
        }


    https://icinga.com/docs/icinga-2/latest/doc/09-object-types/#eventcommand
    :see: `doc/09-object-types.md L295-L320 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L295-L320>`__

    .. tags:: Object type, Monitoring object type
    """


@dataclass(config={"extra": "forbid"})
class Host(Checkable):
    """
    A host.

    .. code-block::

        object Host "icinga2-agent1.localdomain" {
            display_name = "Linux Client 1"
            address = "192.168.56.111"
            address6 = "2a00:1450:4001:815::2003"

            groups = [ "linux-servers" ]

            check_command = "hostalive"
        }

    :see: `Icinga2 documentation: Host <https://icinga.com/docs/icinga-2/latest/doc/09-object-types/#host>`__
    :see: `doc/09-object-types.md L323-L413 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L323-L413>`__
    :see: `lib/icinga/host.ti <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/host.ti>`__

    .. tags:: Object type, Monitoring object type
    """

    groups: Optional[Sequence[str]] = None
    """

     A list of host groups this host belongs to.

     :see: `lib/icinga/host.ti L18-L20 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/host.ti#L18-L20>`__
    """

    display_name: Optional[str] = None
    """
    A short description of the host (e.g. displayed by external interfaces instead of the name if set).

    :see: `lib/icinga/host.ti L22-L30 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/host.ti#L22-L30>`__
    """

    address: Optional[str] = None
    """
    The host's IPv4 address. Available as command runtime macro ``$address$`` if set.

    :see: `lib/icinga/host.ti L32 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/host.ti#L32>`__
    """

    address6: Optional[str] = None
    """
    The host's IPv6 address. Available as command runtime macro ``$address6$`` if set.

    :see: `lib/icinga/host.ti L33 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/host.ti#L33>`__
    """

    state: Optional[HostState] = None
    """

    The current state (0 = UP, 1 = DOWN).

    :see: `lib/icinga/host.ti L35-L37 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/host.ti#L35-L37>`__
    """

    last_state: Optional[HostState] = None
    """
    The previous state (0 = UP, 1 = DOWN).

    :see: `lib/icinga/host.ti L38-L40 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/host.ti#L38-L40>`__
    """

    last_hard_state: Optional[HostState] = None
    """
    The last hard state (0 = UP, 1 = DOWN).

    :see: `lib/icinga/host.ti L41-L43 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/host.ti#L41-L43>`__
    """

    last_state_up: Optional[Timestamp] = None
    """
    When the last UP state occurred (as a UNIX timestamp).

    :see: `lib/icinga/host.ti L44 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/host.ti#L44>`__
    """

    last_state_down: Optional[Timestamp] = None
    """
    When the last DOWN state occurred (as a UNIX timestamp).

    :see: `lib/icinga/host.ti L45 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/host.ti#L45>`__
    """


@dataclass(config={"extra": "forbid"})
class HostGroup:
    """
    A group of hosts.

    .. code-block::

        object HostGroup "linux-servers" {
            display_name = "Linux Servers"

            assign where host.vars.os == "Linux"
        }

    https://icinga.com/docs/icinga-2/latest/doc/09-object-types/#hostgroup
    :see: `doc/09-object-types.md L417-L440 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L417-L440>`__

    .. tags:: Object type, Monitoring object type
    """


@dataclass(config={"extra": "forbid"})
class Notification:
    """
    Notification objects are used to specify how users should be notified in case
    of host and service state changes and other events.

    Example:

    .. code-block::

        object Notification "localhost-ping-notification" {
            host_name = "localhost"
            service_name = "ping4"

            command = "mail-notification"

            users = [ "user1", "user2" ] // reference to User objects

            types = [ Problem, Recovery ]
            states = [ Critical, Warning, OK ]
        }

    https://icinga.com/docs/icinga-2/latest/doc/09-object-types/#notification
    :see: `doc/09-object-types.md L444-L527 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L444-L527>`__

    .. tags:: Object type, Monitoring object type
    """


@dataclass(config={"extra": "forbid"})
class NotificationCommand:
    """
    A notification command definition.

    .. code-block::

        object NotificationCommand "mail-service-notification" {
            command = [ ConfigDir + "/scripts/mail-service-notification.sh" ]

            arguments += {
                "-4" = {
                    required = true
                    value = "$notification_address$"
                }
                "-6" = "$notification_address6$"
                "-b" = "$notification_author$"
                "-c" = "$notification_comment$"
                "-d" = {
                    required = true
                    value = "$notification_date$"
                }
                "-e" = {
                    required = true
                    value = "$notification_servicename$"
                }
                "-f" = {
                    value = "$notification_from$"
                    description = "Set from address. Requires GNU mailutils (Debian/Ubuntu) or mailx (RHEL/SUSE)"
                }
                "-i" = "$notification_icingaweb2url$"
                "-l" = {
                    required = true
                    value = "$notification_hostname$"
                }
                "-n" = {
                    required = true
                    value = "$notification_hostdisplayname$"
                }
                "-o" = {
                    required = true
                    value = "$notification_serviceoutput$"
                }
                "-r" = {
                    required = true
                    value = "$notification_useremail$"
                }
                "-s" = {
                    required = true
                    value = "$notification_servicestate$"
                }
                "-t" = {
                    required = true
                    value = "$notification_type$"
                }
                "-u" = {
                    required = true
                    value = "$notification_servicedisplayname$"
                }
                "-v" = "$notification_logtosyslog$"
            }

            vars += {
                notification_address = "$address$"
                notification_address6 = "$address6$"
                notification_author = "$notification.author$"
                notification_comment = "$notification.comment$"
                notification_type = "$notification.type$"
                notification_date = "$icinga.long_date_time$"
                notification_hostname = "$host.name$"
                notification_hostdisplayname = "$host.display_name$"
                notification_servicename = "$service.name$"
                notification_serviceoutput = "$service.output$"
                notification_servicestate = "$service.state$"
                notification_useremail = "$user.email$"
                notification_servicedisplayname = "$service.display_name$"
            }
        }

    https://icinga.com/docs/icinga-2/latest/doc/09-object-types/#notificationcommand
    :see: `doc/09-object-types.md L530-L622 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L530-L622>`__

    .. tags:: Object type, Monitoring object type
    """


@dataclass(config={"extra": "forbid"})
class ScheduledDowntime:
    """
    ScheduledDowntime objects can be used to set up recurring downtimes for hosts/services.

    Example:

    .. code-block::


        object ScheduledDowntime "some-downtime" {
            host_name = "localhost"
            service_name = "ping4"

            author = "icingaadmin"
            comment = "Some comment"

            fixed = false
            duration = 30m

            ranges = {
                "sunday" = "02:00-03:00"
            }
        }

    https://icinga.com/docs/icinga-2/latest/doc/09-object-types/#scheduleddowntime
    :see: `doc/09-object-types.md L624-L674 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L624-L674>`__

    .. tags:: Object type, Monitoring object type
    """


@dataclass(config={"extra": "forbid"})
class Service(Checkable):
    """
    Service objects describe network services and how they should be checked
    by Icinga 2.

    Best Practice

    Rather than creating a ``Service`` object for a specific host it is usually easier
    to just create a ``Service`` template and use the ``apply`` keyword to assign the
    service to a float of hosts.
    Check the `apply <03-monitoring-basics.md#using-apply>`__ chapter for details.

    Example:

    .. code-block::


        object Service "uptime" {
            host_name = "localhost"

            display_name = "localhost Uptime"

            check_command = "snmp"

            vars.snmp_community = "public"
            vars.snmp_oid = "DISMAN-EVENT-MIB::sysUpTimeInstance"

            check_interval = 60s
            retry_interval = 15s

            groups = [ "all-services", "snmp" ]
        }

    .. tags:: Object type, Monitoring object type

    https://icinga.com/docs/icinga-2/latest/doc/09-object-types/#service
    :see: `lib/icinga/service.ti <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/service.ti>`__
    :see: `doc/09-object-types.md L677-L781 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L677-L781>`__
    """

    groups: Optional[Sequence[str]] = None
    """
    The service groups this service belongs to.
    """

    display_name: Optional[str] = None
    """
    A short description of the service.

    :see: `doc/09-object-types.md L712 <https://github.com/Icinga/icinga2/blob/0951230ce1be27c9957ef8801be258524524dc01/doc/09-object-types.md?plain=1#L712>`__
    :see: `lib/icinga/service.ti L34-L42 <https://github.com/Icinga/icinga2/blob/0951230ce1be27c9957ef8801be258524524dc01/lib/icinga/service.ti#L34-L42>`__
    """

    host_name: Optional[str] = None
    """
    The host this service belongs to. There must be a `Host` object with that name.
    """

    host: Optional[Host] = None

    state: Optional[ServiceState] = None
    """
    The current state (0 = OK, 1 = WARNING, 2 = CRITICAL, 3 = UNKNOWN).
    """

    last_state: Optional[ServiceState] = None
    """
    The previous state (0 = OK, 1 = WARNING, 2 = CRITICAL, 3 = UNKNOWN).
    """

    last_hard_state: Optional[ServiceState] = None
    """
    The last hard state (0 = OK, 1 = WARNING, 2 = CRITICAL, 3 = UNKNOWN).
    """

    last_state_ok: Optional[Timestamp] = None
    """
    When the last OK state occurred (as a UNIX timestamp).
    """

    last_state_warning: Optional[Timestamp] = None
    """
    When the last WARNING state occurred (as a UNIX timestamp).
    """

    last_state_critical: Optional[Timestamp] = None
    """
    When the last CRITICAL state occurred (as a UNIX timestamp).
    """

    last_state_unknown: Optional[Timestamp] = None
    """
    When the last UNKNOWN state occurred (as a UNIX timestamp).
    """


@dataclass(config={"extra": "forbid"})
class ServiceGroup:
    """
    A group of services.

    Example:

    .. code-block::

        object ServiceGroup "snmp" {
            display_name = "SNMP services"
        }

    https://icinga.com/docs/icinga-2/latest/doc/09-object-types/#servicegroup
    :see: `doc/09-object-types.md L784-L805 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L784-L805>`__

    .. tags:: Object type, Monitoring object type
    """


@dataclass(config={"extra": "forbid"})
class TimePeriodSegment:
    """
    :see: `lib/icinga/timeperiod.cpp L84-L87 <https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/icinga/timeperiod.cpp#L84-L87>`__
    """

    begin: int

    end: int


@dataclass(config={"extra": "forbid"})
class TimePeriod(CustomVarObject):
    """
    Time periods can be used to specify when hosts/services should be checked or to limit
    when notifications should be sent out.

    Examples:

    .. code-block::

        object TimePeriod "nonworkhours" {
            display_name = "Icinga 2 TimePeriod for non working hours"

            ranges = {
                monday = "00:00-8:00,17:00-24:00"
                tuesday = "00:00-8:00,17:00-24:00"
                wednesday = "00:00-8:00,17:00-24:00"
                thursday = "00:00-8:00,17:00-24:00"
                friday = "00:00-8:00,16:00-24:00"
                saturday = "00:00-24:00"
                sunday = "00:00-24:00"
            }
        }

    .. code-block::

        object TimePeriod "exampledays" {
            display_name = "Icinga 2 TimePeriod for random example days"

            ranges = {
                //We still believe in Santa, no peeking!
                //Applies every 25th of December every year
                "december 25" = "00:00-24:00"

                //Any point in time can be specified,
                //but you still have to use a range
                "2038-01-19" = "03:13-03:15"

                //Evey 3rd day from the second monday of February
                //to 8th of November
                "monday 2 february - november 8 / 3" = "00:00-24:00"
            }
        }

    Additional examples can be found `here <08-advanced-topics.md#timeperiods>`__.

    .. tags:: Object type, Monitoring object type

    https://icinga.com/docs/icinga-2/latest/doc/09-object-types/#timeperiod
    :see: `doc/09-object-types.md L809-L869 <https://github.com/Icinga/icinga2/blob/4c6b93d61775ff98fc671b05ad4de2b62945ba6a/doc/09-object-types.md?plain=1#L807-L867>`__
    :see: `lib/icinga/timeperiod.ti L11-L39 <https://github.com/Icinga/icinga2/blob/894d6aa290e83797d001fcc2887611b23707dbf9/lib/icinga/timeperiod.ti#L11-L39>`__
    """

    display_name: Optional[str] = None
    """
    A short description of the time period.

    :see: `lib/icinga/timeperiod.ti L13-L21 <https://github.com/Icinga/icinga2/blob/894d6aa290e83797d001fcc2887611b23707dbf9/lib/icinga/timeperiod.ti#L13-L21>`__
    """

    ranges: Optional[dict[str, str]] = None
    """
    A dictionary containing information which days and durations apply to this timeperiod.

    :see: `lib/icinga/timeperiod.ti L22 <https://github.com/Icinga/icinga2/blob/894d6aa290e83797d001fcc2887611b23707dbf9/lib/icinga/timeperiod.ti#L22>`__
    """

    update: Optional[Payload] = None
    """
    :see: `lib/icinga/timeperiod.ti L23 <https://github.com/Icinga/icinga2/blob/894d6aa290e83797d001fcc2887611b23707dbf9/lib/icinga/timeperiod.ti#L23>`__
    """

    prefer_includes: Optional[bool] = None
    """
    Whether to prefer timeperiods ``includes`` or ``excludes``. Default to true.

    :see: `lib/icinga/timeperiod.ti L24-L26 <https://github.com/Icinga/icinga2/blob/894d6aa290e83797d001fcc2887611b23707dbf9/lib/icinga/timeperiod.ti#L24-L26>`__
    """

    excludes: Optional[Sequence[str]] = None
    """
    An array of timeperiods, which should exclude from your timerange.

    :see: `lib/icinga/timeperiod.ti L27-L29 <https://github.com/Icinga/icinga2/blob/894d6aa290e83797d001fcc2887611b23707dbf9/lib/icinga/timeperiod.ti#L27-L29>`__
    """

    includes: Optional[Sequence[str]] = None
    """
    An array of timeperiods, which should include into your timerange.

    :see: `lib/icinga/timeperiod.ti L30-L32 <https://github.com/Icinga/icinga2/blob/894d6aa290e83797d001fcc2887611b23707dbf9/lib/icinga/timeperiod.ti#L30-L32>`__
    """

    valid_begin: Optional[Timestamp] = None
    """
    :see: `lib/icinga/timeperiod.ti L33 <https://github.com/Icinga/icinga2/blob/894d6aa290e83797d001fcc2887611b23707dbf9/lib/icinga/timeperiod.ti#L33>`__
    """

    valid_end: Optional[Timestamp] = None
    """
    :see: `lib/icinga/timeperiod.ti L34 <https://github.com/Icinga/icinga2/blob/894d6aa290e83797d001fcc2887611b23707dbf9/lib/icinga/timeperiod.ti#L34>`__
    """

    segments: Optional[Sequence[TimePeriodSegment]] = None
    """
    :see: `lib/icinga/timeperiod.ti L35 <https://github.com/Icinga/icinga2/blob/894d6aa290e83797d001fcc2887611b23707dbf9/lib/icinga/timeperiod.ti#L35>`__
    """

    is_inside: Optional[bool] = None
    """
    :see: `lib/icinga/timeperiod.ti L36-L38 <https://github.com/Icinga/icinga2/blob/894d6aa290e83797d001fcc2887611b23707dbf9/lib/icinga/timeperiod.ti#L36-L38>`__
    """


@dataclass(config={"extra": "forbid"})
class User(CustomVarObject):
    """A user.

    Example:

    .. code-block::

        object User "icingaadmin" {
            display_name = "Icinga 2 Admin"
            groups = [ "icingaadmins" ]
            email = "icinga@localhost"
            pager = "icingaadmin@localhost.localdomain"

            period = "24x7"

            states = [ OK, Warning, Critical, Unknown ]
            types = [ Problem, Recovery ]

            vars.additional_notes = "This is the Icinga 2 Admin account."
        }


    Available notification state filters:

    .. code-block::

        OK
        Warning
        Critical
        Unknown
        Up
        Down


    Available notification type filters:

    .. code-block::

        DowntimeStart
        DowntimeEnd
        DowntimeRemoved
        Custom
        Acknowledgement
        Problem
        Recovery
        FlappingStart
        FlappingEnd

    .. tags:: Object type, Monitoring object type

    https://icinga.com/docs/icinga-2/latest/doc/09-object-types/#user
    :see: `lib/icinga/user.ti <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/user.ti>`__
    :see: `doc/09-object-types.md L872-L937 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L872-L937>`__
    """

    display_name: Optional[str] = None
    """
    A short description of the user.

    :see: `lib/icinga/user.ti L14-L22 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/user.ti#L14-L22>`__
    :see: `doc/09-object-types.md L923 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L923>`__

    """

    groups: Optional[Sequence[str]] = None
    """
    An array of group names.

    :see: `lib/icinga/user.ti L23-L25 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/user.ti#L23-L25>`__
    :see: `doc/09-object-types.md L927 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L927>`__

    """

    period: Optional[str] = None
    """
    The name of a time period which determines when a notification for this user should be triggered. Not set by default (effectively 24x7).

    :see: `lib/icinga/user.ti L26-L30 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/user.ti#L26-L30>`__
    :see: `doc/09-object-types.md L929 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L929>`__

    """

    types: Optional[Sequence[str]] = None
    """
    A set of type filters when a notification for this user should be triggered. By default everything is matched.

    :see: `lib/icinga/user.ti L32 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/user.ti#L32>`__
    :see: `doc/09-object-types.md L930 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L930>`__

    """

    states: Optional[Sequence[str]] = None
    """
    A set of state filters when a notification for this should be triggered. By default everything is matched.

    :see: `lib/icinga/user.ti L34 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/user.ti#L34>`__
    :see: `doc/09-object-types.md L931 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L931>`__

    """

    email: Optional[str] = None
    """
    An email string for this user. Useful for notification commands.

    :see: `lib/icinga/user.ti L37 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/user.ti#L37>`__
    :see: `doc/09-object-types.md L924 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L924>`__

    """

    pager: Optional[str] = None
    """
    A pager str for this user. Useful for notification commands.

    :see: `lib/icinga/user.ti L38 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/user.ti#L38>`__
    :see: `doc/09-object-types.md L925 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L925>`__

    """

    enable_notifications: Optional[bool] = None
    """
    Whether notifications are enabled for this user. Defaults to true.

    :see: `lib/icinga/user.ti L40-L42 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/user.ti#L40-L42>`__
    :see: `doc/09-object-types.md L928 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L928>`__

    """

    last_notification: Optional[float] = None
    """
    When the last notification was sent for this user (as a UNIX timestamp).

    :see: `lib/icinga/user.ti L44 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/lib/icinga/user.ti#L44>`__
    :see: `doc/09-object-types.md L937 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L937>`__

    """


@dataclass(config={"extra": "forbid"})
class UserGroup(CustomVarObject):
    """
    A user group.

    Example:

    .. code-block::

        object UserGroup "icingaadmins" {
            display_name = "Icinga 2 Admin Group"
        }

    .. tags:: Object type, Monitoring object type

    https://icinga.com/docs/icinga-2/latest/doc/09-object-types/#usergroup
    :see: `doc/09-object-types.md L939-L960 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L939-L960>`__
    https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/icinga/usergroup.ti#L10-L23
    """

    display_name: OptionalStr = None
    """Optional. A short description of the user group."""

    groups: Optional[Sequence[str]] = None
    """Optional. An array of nested group names."""


@dataclass(config={"extra": "forbid"})
class Zone(ConfigObject):
    """
    Zone objects are used to specify which Icinga 2 instances are located in a zone.

    Example:

    .. code-block::

        object Zone "master" {
            endpoints = [ "icinga2-master1.localdomain", "icinga2-master2.localdomain" ]
        }

        object Zone "satellite" {
            endpoints = [ "icinga2-satellite1.localdomain" ]
            parent = "master"
        }

    .. tags:: Object type, Monitoring object type

    https://icinga.com/docs/icinga-2/latest/doc/09-object-types/#zone
    :see: `doc/09-object-types.md L963-L989 <https://github.com/Icinga/icinga2/blob/2c9117b4f71e00b2072e7dbe6c4ea4e48c882a87/doc/09-object-types.md?plain=1#L963-L989>`__
    https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/remote/zone.ti#L10-L23
    """

    endpoints: Optional[Sequence[str]] = None
    """Optional. Array of endpoint names located in this zone."""

    parent: OptionalStr = None
    """Optional. The name of the parent zone. (Do not specify a global zone)"""

    all_parents: Optional[Sequence[str]] = None

    is_global: Optional[bool] = None
    """Optional. Whether configuration files for this zone should be synced to all endpoints. Defaults to false."""


########################################################################################
# Runtime Objects
########################################################################################


@dataclass(config={"extra": "forbid"})
class Comment:
    """
    .. tags:: Object type, Runtime object type
    """


@dataclass(config={"extra": "forbid"})
class Downtime:
    """
    Downtimes created at runtime are represented as objects. You can create
    downtimes with the schedule-downtime API action.

    Example:

    .. code-block::

        object Downtime "my-downtime" {
            host_name = "localhost"
            author = "icingaadmin"
            comment = "This is a downtime."
            start_time = 1505312869
            end_time = 1505312924
        }


    .. tags:: Object type, Runtime object type

    :see: `doc/09-object-types/#downtime <https://icinga.com/docs/icinga-2/latest/doc/09-object-types/#downtime>`__
    """

    host_name: Optional[str] = None
    """Required. The name of the host this downtime belongs to."""

    service_name: Optional[str] = None
    """Optional. The short name of the service this downtime belongs to. If
    omitted, this downtime object is treated as host downtime."""

    author: Optional[str] = None
    """Required. The author’s name."""

    comment: Optional[str] = None
    """Required. The comment text."""

    start_time: Optional[Timestamp] = None
    """ 	Required. The start time as UNIX timestamp."""

    end_time: Optional[Timestamp] = None
    """Timestamp 	Required. The end time as UNIX timestamp."""

    duration: Optional[int] = None
    """Number 	Optional. The duration as number."""

    entry_time: Optional[Timestamp] = None
    """Timestamp 	Optional. The UNIX timestamp when this downtime was
    added."""

    fixed: Optional[bool] = None
    """Boolean 	Optional. Whether the downtime is fixed (true) or flexible
    (false). Defaults to flexible. Details in the advanced topics chapter."""

    triggers: Optional[Sequence[str]] = None
    """Array of object names"""


########################################################################################
# Features
########################################################################################


@dataclass(config={"extra": "forbid"})
class ApiListener:
    """
    .. tags:: Object type, Feature object type
    """


@dataclass(config={"extra": "forbid"})
class CheckerComponent:
    """
    .. tags:: Object type, Feature object type
    """


@dataclass(config={"extra": "forbid"})
class CompatLogger:
    """
    .. tags:: Object type, Feature object type
    """


@dataclass(config={"extra": "forbid"})
class ElasticsearchWriter:
    """
    .. tags:: Object type, Feature object type
    """


@dataclass(config={"extra": "forbid"})
class ExternalCommandListener:
    """
    .. tags:: Object type, Feature object type
    """


@dataclass(config={"extra": "forbid"})
class FileLogger:
    """
    .. tags:: Object type, Feature object type
    """


@dataclass(config={"extra": "forbid"})
class GelfWriter:
    """
    .. tags:: Object type, Feature object type
    """


@dataclass(config={"extra": "forbid"})
class GraphiteWriter:
    """
    .. tags:: Object type, Feature object type
    """


@dataclass(config={"extra": "forbid"})
class IcingaApplication:
    """
    .. tags:: Object type, Feature object type
    """


@dataclass(config={"extra": "forbid"})
class IcingaDB:
    """
    .. tags:: Object type, Feature object type
    """


@dataclass(config={"extra": "forbid"})
class IdoMySqlConnection:
    """
    .. tags:: Object type, Feature object type
    """


@dataclass(config={"extra": "forbid"})
class IdoPgsqlConnection:
    """
    .. tags:: Object type, Feature object type
    """


@dataclass(config={"extra": "forbid"})
class InfluxdbWriter:
    """
    .. tags:: Object type, Feature object type
    """


@dataclass(config={"extra": "forbid"})
class Influxdb2Writer:
    """
    .. tags:: Object type, Feature object type
    """


@dataclass(config={"extra": "forbid"})
class JournaldLogger:
    """
    .. tags:: Object type, Feature object type
    """


@dataclass(config={"extra": "forbid"})
class LiveStatusListener:
    """
    .. tags:: Object type, Feature object type
    """


@dataclass(config={"extra": "forbid"})
class NotificationComponent:
    """
    .. tags:: Object type, Feature object type
    """


@dataclass(config={"extra": "forbid"})
class OpenTsdbWriter:
    """
    .. tags:: Object type, Feature object type
    """


@dataclass(config={"extra": "forbid"})
class PerfdataWriter:
    """
    .. tags:: Object type, Feature object type
    """


@dataclass(config={"extra": "forbid"})
class SyslogLogger:
    """
    .. tags:: Object type, Feature object type
    """


@dataclass(config={"extra": "forbid"})
class WindowsEventLogLogger:
    """
    .. tags:: Object type, Feature object type
    """


@dataclass
class EventStreamTypeCheckResult:
    """https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#event-stream-type-checkresult"""

    type: Literal["CheckResult"]

    timestamp: Timestamp
    """Unix timestamp when the event happened."""

    host: str
    """Host name."""

    check_result: CheckResult
    """Serialized CheckResult value type."""

    downtime_depth: float
    """Amount of active downtimes on the checkable."""

    acknowledgement: bool
    """Whether the object is acknowledged."""

    service: Optional[str] = None
    """Service name. Optional if this is a host check result."""


@dataclass
class EventStreamTypeStateChange:
    """https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#event-stream-type-statechange"""

    type: Literal["StateChange"]


@dataclass
class EventStreamTypeNotification:
    """https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#event-stream-type-notification"""

    type: Literal["Notification"]


@dataclass
class EventStreamTypeFlapping:
    """https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#event-stream-type-flapping"""

    type: Literal["Flapping"]


@dataclass
class EventStreamTypeAcknowledgementSet:
    """https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#event-stream-type-acknowledgementset"""

    type: Literal["AcknowledgementSet"]


@dataclass
class EventStreamTypeAcknowledgementCleared:
    """https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#event-stream-type-acknowledgementcleared"""

    type: Literal["AcknowledgementCleared"]


@dataclass
class _EventStreamTypeComment:
    timestamp: Timestamp
    """Unix timestamp when the event happened."""

    comment: Comment
    """Serialized Comment object."""


@dataclass
class EventStreamTypeCommentAdded(_EventStreamTypeComment):
    """https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#event-stream-type-commentadded"""

    type: Literal["CommentAdded"]


@dataclass
class EventStreamTypeCommentRemoved(_EventStreamTypeComment):
    """https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#event-stream-type-commentremoved"""

    type: Literal["CommentRemoved"]


@dataclass
class _EventStreamTypeDowntime:
    timestamp: Timestamp
    """Unix timestamp when the event happened."""

    downtime: Downtime
    """Serialized Downtime object."""


@dataclass
class EventStreamTypeDowntimeAdded(_EventStreamTypeDowntime):
    """https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#event-stream-type-downtimeadded"""

    type: Literal["DowntimeAdded"]


@dataclass
class EventStreamTypeDowntimeRemoved(_EventStreamTypeDowntime):
    """https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#event-stream-type-downtimeremoved"""

    type: Literal["DowntimeRemoved"]


@dataclass
class EventStreamTypeDowntimeStarted(_EventStreamTypeDowntime):
    """https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#event-stream-type-downtimestarted"""

    type: Literal["DowntimeStarted"]


@dataclass
class EventStreamTypeDowntimeTriggered(_EventStreamTypeDowntime):
    """https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#event-stream-type-downtimetriggered"""

    type: Literal["DowntimeTriggered"]


EventStream = Union[
    EventStreamTypeCheckResult,
    EventStreamTypeDowntimeAdded,
    EventStreamTypeDowntimeStarted,
    EventStreamTypeDowntimeRemoved,
    EventStreamTypeDowntimeTriggered,
]



# Copyright 2017 fmnisme@gmail.com christian@jonak.org
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# @author: Christian Jonak-Moechel, fmnisme, Tobias von der Krone
# @contact: christian@jonak.org, fmnisme@gmail.com, tobias@vonderkrone.info
# @summary: Python library for the Icinga 2 RESTful API

"""
Icinga 2 API client

The Icinga 2 API allows you to manage configuration objects and resources in a simple,
programmatic way using HTTP requests.
"""

import urllib
import urllib.parse
from collections.abc import Sequence
from importlib.metadata import version as get_version
from typing import Any, Generator, Literal, Optional, Union

from pydantic.dataclasses import dataclass

from pretiac.config import Config
from pretiac.exceptions import PretiacException
from pretiac.object_types import (
    CheckResult,
    FilterVars,
    HostOrService,
    ObjectTypeName,
    Payload,
    Value,
    pluralize_to_lower_object_type_name,
)
from pretiac.request_handler import (
    RequestHandler,
    State,
    normalize_state,
)


def assemble_payload(**kwargs: Any) -> Payload:
    # https://stackoverflow.com/a/2544761
    return {k: v for k, v in kwargs.items() if v is not None}


@dataclass
class Result:
    code: int

    status: str


@dataclass
class ResultContainer:
    results: list[Result]


class ActionsUrlEndpoint(RequestHandler):
    """
    Connects to the URL endpoint ``actions`` of the Icinga2 API.
    """

    path_prefix = "actions"

    def process_check_result(
        self,
        type: HostOrService,
        name: str,
        exit_status: State,
        plugin_output: str,
        performance_data: Optional[Sequence[str] | str] = None,
        check_command: Optional[Sequence[str] | str] = None,
        check_source: Optional[str] = None,
        execution_start: Optional[float] = None,
        execution_end: Optional[float] = None,
        ttl: Optional[int] = None,
        filter: Optional[str] = None,
        filter_vars: FilterVars = None,
        suppress_exception: Optional[bool] = None,
    ) -> Any:
        """Process a check result for a host or a service.

        Send a ``POST`` request to the URL endpoint ``actions/process-check-result``.

        :param type: ``Host`` or ``Service``.
        :param name: The name of the object.
        :param exit_status: For services: ``0=OK``, ``1=WARNING``, ``2=CRITICAL``,
            ``3=UNKNOWN``, for hosts: ``0=UP``, ``1=DOWN``.
        :param plugin_output: One or more lines of the plugin main output. Does **not**
            contain the performance data.
        :param check_command: The first entry should be the check commands path, then
            one entry for each command line option followed by an entry for each of its
            argument. Alternativly a single string can be used.
        :param check_source: Usually the name of the ``command_endpoint``.
        :param execution_start: The timestamp where a script/process started its
            execution.
        :param execution_end: The timestamp where a script/process ended its execution.
            This timestamp is used in features to determine e.g. the metric timestamp.
        :param ttl: Time-to-live duration in seconds for this check result. The next
            expected check result is ``now + ttl`` where freshness checks are executed.
        :param filter: filters matched object(s)
        :param filter_vars: variables used in the filters expression
        :param suppress_exception: If this parameter is set to ``True``, no exceptions
            are thrown.

        :returns: the response as json

        .. code-block:: python

            raw_client.process_check_result(
                "Service",
                "myhost.domain!ping4",
                exit_status=2,
                plugin_output="PING CRITICAL - Packet loss = 100%",
                performance_data=[
                    "rta=5000.000000ms;3000.000000;5000.000000;0.000000",
                    "pl=100%;80;100;0",
                ],
                check_source="python client",
            )

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#process-check-result <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#process-check-result>`__
        """
        if not name and not filter:
            raise PretiacException("name and filters is empty or none")

        if name and (filter or filter_vars):
            raise PretiacException("name and filters are mutually exclusive")

        if type not in ["Host", "Service"]:
            raise PretiacException('type needs to be "Host" or "Service".')

        payload: Payload = {
            "type": type,
            "exit_status": normalize_state(exit_status),
            "plugin_output": plugin_output,
        }

        if name:
            payload[type.lower()] = name
        if performance_data:
            payload["performance_data"] = performance_data
        if check_command:
            payload["check_command"] = check_command
        if check_source:
            payload["check_source"] = check_source
        if execution_start:
            payload["execution_start"] = execution_start
        if execution_end:
            payload["execution_end"] = execution_end
        if ttl:
            payload["ttl"] = ttl
        if filter:
            payload["filter"] = filter
        if filter_vars:
            payload["filter_vars"] = filter_vars

        return self._request(
            "POST",
            "process-check-result",
            payload,
            suppress_exception=suppress_exception,
        )

    def reschedule_check(
        self,
        object_type: HostOrService,
        filters: str,
        filter_vars: FilterVars = None,
        next_check: Optional[int] = None,
        force_check: Optional[bool] = True,
    ) -> Any:
        """
        Reschedule a check for hosts and services.

        example 1:

        .. code-block:: python

            raw_client.reschedule_check("Service", 'service.name=="ping4"')

        example 2:

        .. code-block:: python

            raw_client.reschedule_check("Host", 'host.name=="localhost"', 1577833200)

        :param object_type: Host or Service
        :param filters: filters matched object(s)
        :param filter_vars: variables used in the for filters expression
        :param next_check: timestamp to run the check
        :param force: ignore period restrictions and disabled checks

        :returns: the response as json

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#reschedule-check <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#reschedule-check>`__
        """

        payload: Payload = {
            "type": object_type,
            "filter": filters,
            "force_check": force_check,
        }
        if next_check:
            payload["next_check"] = next_check
        if filter_vars:
            payload["filter_vars"] = filter_vars

        return self._request("POST", "reschedule-check", payload)

    def send_custom_notification(
        self,
        object_type: HostOrService,
        filters: str,
        author: str,
        comment: str,
        filter_vars: FilterVars = None,
        force: Optional[int] = False,
    ) -> Any:
        """
        Send a custom notification for hosts and services.

        example 1:

        .. code-block:: python

            send_custom_notification(
                "Host", "host.name==localhost", "icingaadmin", "test comment"
            )

        :param object_type: Host or Service
        :param filters: filters matched object
        :param author: name of the author
        :param comment: comment text
        :param force: ignore downtimes and notification settings
        :param filter_vars: variables used in the filters expression

        :returns: the response as json

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#send-custom-notification <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#send-custom-notification>`__
        """
        payload: Payload = {
            "type": object_type,
            "filter": filters,
            "author": author,
            "comment": comment,
            "force": force,
        }
        if filter_vars:
            payload["filter_vars"] = filter_vars

        return self._request("POST", "send-custom-notification", payload)

    def delay_notification(
        self,
        object_type: HostOrService,
        filters: str,
        timestamp: int,
        filter_vars: FilterVars = None,
    ) -> Any:
        """
        Delay notifications for a host or a service.

        example 1:

        .. code-block:: python

            delay_notification("Service", "1446389894")

            delay_notification("Host", 'host.name=="localhost"', "1446389894")

        :param object_type: Host or Service
        :param filters: filters matched object(s)
        :param timestamp: timestamp to delay the notifications to
        :param filter_vars: variables used in the filters expression

        :returns: the response as json

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#delay-notification <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#delay-notification>`__
        """
        payload: Payload = {
            "type": object_type,
            "filter": filters,
            "timestamp": timestamp,
        }
        if filter_vars:
            payload["filter_vars"] = filter_vars

        return self._request("POST", "delay-notification", payload)

    def acknowledge_problem(
        self,
        object_type: HostOrService,
        filters: str,
        author: str,
        comment: str,
        filter_vars: FilterVars = None,
        expiry: Optional[int] = None,
        sticky: Optional[bool] = None,
        notify: Optional[bool] = None,
        persistent: Optional[bool] = None,
    ) -> Any:
        """
        Acknowledge a Service or Host problem.

        :param object_type: Host or Service
        :param filters: filters matched object(s)
        :param author: name of the author
        :param comment: comment text
        :param filter_vars: variables used in the filters expression
        :param expiry: acknowledgement expiry timestamp
        :param sticky: stay till full problem recovery
        :param notify: send notification
        :param persistent: the comment will remain after the acknowledgement recovers or expires

        :returns: the response as json

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#acknowledge-problem <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#acknowledge-problem>`__
        """
        payload: Payload = {
            "type": object_type,
            "filter": filters,
            "author": author,
            "comment": comment,
        }
        if filter_vars:
            payload["filter_vars"] = filter_vars
        if expiry:
            payload["expiry"] = expiry
        if sticky:
            payload["sticky"] = sticky
        if notify:
            payload["notify"] = notify
        if persistent:
            payload["persistent"] = persistent

        return self._request("POST", "acknowledge-problem", payload)

    def remove_acknowledgement(
        self, object_type: HostOrService, filters: str, filter_vars: FilterVars = None
    ) -> Any:
        """
        Remove the acknowledgement for services or hosts.

        example 1:

        .. code-block:: python

            raw_client.actions.remove_acknowledgement(
                object_type="Service", filters="service.state==2"
            )

        :param object_type: Host or Service
        :param filters: filters matched object(s)
        :param filter_vars: variables used in the filters expression

        :returns: the response as json

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#remove-acknowledgement <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#remove-acknowledgement>`__
        """
        payload: Payload = {"type": object_type, "filter": filters}
        if filter_vars:
            payload["filter_vars"] = filter_vars

        return self._request("POST", "remove-acknowledgement", payload)

    def add_comment(
        self,
        object_type: HostOrService,
        filter: str,
        author: str,
        comment: str,
        filter_vars: FilterVars = None,
    ) -> Any:
        """
        Add a ``comment`` from an ``author`` to services or hosts.

        example 1:

        .. code-block:: python

            raw_client.actions.add_comment(
                "Service",
                'service.name=="ping4"',
                "icingaadmin",
                "Troubleticket #123456789 opened.",
            )

        :param object_type: The valid types for this action are Host and Service.
        :param filter: filters matched object(s)
        :param author: name of the author
        :param comment: comment text
        :param filter_vars: variables used in the filters expression

        :returns: the response as json

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#add-comment <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#add-comment>`__
        """
        return self._request(
            "POST",
            "add-comment",
            assemble_payload(
                type=object_type,
                filter=filter,
                author=author,
                comment=comment,
                filter_vars=filter_vars,
            ),
        )

    def remove_comment(
        self,
        object_type: HostOrService,
        name: Optional[str] = None,
        filters: Optional[str] = None,
        filter_vars: FilterVars = None,
    ) -> Any:
        """
        Remove a comment using its name or filters.

        example 1:

        .. code-block:: python

            remove_comment("Comment" "localhost!localhost-1458202056-25")

        example 2:

        .. code-block:: python

            remove_comment('Service'
                        filters='service.name=="ping4"')

        :param object_type: Host, Service or Comment
        :param name: name of the Comment
        :param filters: filters matched object(s)
        :param filter_vars: variables used in the filters expression

        :returns: the response as json

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#remove-comment <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#remove-comment>`__
        """
        if not name and not filters:
            raise PretiacException("name and filters is empty or none")

        payload: Payload = {"type": object_type}
        if name:
            payload[object_type.lower()] = name
        if filters:
            payload["filter"] = filters
        if filter_vars:
            payload["filter_vars"] = filter_vars

        return self._request("POST", "remove-comment", payload)

    def schedule_downtime(
        self,
        object_type: HostOrService,
        filters: str,
        author: str,
        comment: str,
        start_time: int,
        end_time: int,
        duration: int,
        filter_vars: FilterVars = None,
        fixed: Optional[bool] = None,
        all_services: Optional[bool] = None,
        trigger_name: Optional[str] = None,
        child_options: Optional[str] = None,
    ) -> Any:
        """
        Schedule a downtime for hosts and services.

        example 1:

        .. code-block:: python

            schedule_downtime(
                'object_type': 'Service',
                'filters': r'service.name=="ping4"',
                'author': 'icingaadmin',
                'comment': 'IPv4 network maintenance',
                'start_time': 1446388806,
                'end_time': 1446389806,
                'duration': 1000
            )

        example 2:

        .. code-block:: python

            schedule_downtime(
                'object_type': 'Host',
                'filters': r'match("*", host.name)',
                'author': 'icingaadmin',
                'comment': 'IPv4 network maintenance',
                'start_time': 1446388806,
                'end_time': 1446389806,
                'duration': 1000
            )

        :param object_type: Host or Service
        :param filters: filters matched object(s)
        :param author: name of the author
        :param comment: comment text
        :param start_time: timestamp marking the beginning
        :param end_time: timestamp marking the end
        :param duration: duration of the downtime in seconds
        :param filter_vars: variables used in the filters expression
        :param fixed: fixed or flexible downtime
        :param all_services: sets downtime for all services for the matched host objects
        :param trigger_name: trigger for the downtime
        :param child_options: schedule child downtimes.

        :returns: the response as json

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#schedule-downtime <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#schedule-downtime>`__
        """
        payload: Payload = {
            "type": object_type,
            "filter": filters,
            "author": author,
            "comment": comment,
            "start_time": start_time,
            "end_time": end_time,
            "duration": duration,
        }
        if filter_vars:
            payload["filter_vars"] = filter_vars
        if fixed:
            payload["fixed"] = fixed
        if all_services:
            payload["all_services"] = all_services
        if trigger_name:
            payload["trigger_name"] = trigger_name
        if child_options:
            payload["child_options"] = child_options

        return self._request("POST", "schedule-downtime", payload)

    def remove_downtime(
        self,
        object_type: HostOrService,
        name: Optional[str] = None,
        filters: Optional[str] = None,
        filter_vars: FilterVars = None,
    ) -> Any:
        """
        Remove the downtime using its name or filters.

        example 1:

        .. code-block:: python

            remove_downtime("Downtime", "localhost!ping4!localhost-1458148978-14")

        example 2:

        .. code-block:: python

            remove_downtime("Service", filters='service.name=="ping4"')

        :param object_type: Host, Service or Downtime
        :param name: name of the downtime
        :param filters: filters matched object(s)
        :param filter_vars: variables used in the filters expression

        :returns: the response as json

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#remove-downtime <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#remove-downtime>`__
        """

        if not name and not filters:
            raise PretiacException("name and filters is empty or none")

        payload: Payload = {"type": object_type}
        if name:
            payload[object_type.lower()] = name
        if filters:
            payload["filter"] = filters
        if filter_vars:
            payload["filter_vars"] = filter_vars

        return self._request("POST", "remove-downtime", payload)

    def shutdown_process(self) -> Any:
        """
        Shuts down Icinga2. May or may not return.

        example 1:

        .. code-block:: python

            shutdown_process()

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#shutdown-process <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#shutdown-process>`__
        """
        return self._request("POST", "shutdown-process")

    def restart_process(self) -> Any:
        """
        Restarts Icinga2. May or may not return.

        example 1:

        .. code-block:: python

            restart_process()

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#restart-process <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#restart-process>`__
        """
        return self._request("POST", "restart-process")

    def generate_ticket(self, cn: str) -> Any:
        """
        Generates a PKI ticket for `CSR auto-signing <https://icinga.com/docs/icinga-2/latest/doc/06-distributed-monitoring/#distributed-monitoring-setup-csr-auto-signing>`__.
        This can be used in combination with satellite/client
        setups requesting this ticket number.

        Example:

        .. code-block:: python

            raw_client.actions.generate_ticket("my-server-name")

        :param cn: Required. The host’s common name for which the ticket should be generated.

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#generate-ticket <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#generate-ticket>`__
        """
        return self._request("POST", "generate-ticket", assemble_payload(cn=cn))


class ConfigurationUrlEndpoint(RequestHandler):
    """
    Connects to the URL endpoint ``config`` of the Icinga2 API.
    """

    path_prefix = "config"

    def create_package(
        self,
        package_name: str,
        suppress_exception: Optional[bool] = None,
    ) -> Any:
        """
        Create a new empty configuration package.

        :param package_name: Package names with the ``_`` prefix are reserved for
            internal packages and must not be used. You can recognize
            ``_api``, ``_etc`` and ``_cluster`` when querying specific objects
            and packages.
        :param suppress_exception: If this parameter is set to ``True``, no
            exceptions are thrown.

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#create-a-config-package <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#create-a-config-package>`__
        """
        return self._request(
            "POST",
            f"packages/{_normalize_name(package_name)}",
            suppress_exception=suppress_exception,
        )

    def create_stage(
        self,
        package_name: str,
        files: dict[str, str],
        reload: Optional[bool] = None,
        activate: Optional[bool] = None,
        suppress_exception: Optional[bool] = None,
    ) -> Any:
        """
        Configuration files in packages are managed in stages. Stages provide
        a way to maintain multiple configuration versions for a package.
        Once a new stage is deployed, the content is validated and set as
        active stage on success.

        On failure, the older stage remains active, and the caller can fetch
        the startup.log from this stage deployment attempt to see what exactly
        failed. You can see that in the Director’s deployment log.

        :param package_name: Package names with the ``_`` prefix are reserved for
            internal packages and must not be used. You can recognize
            ``_api``, ``_etc`` and ``_cluster`` when querying specific objects
            and packages.
        :param files: Dictionary of file targets and their content.
        :param reload: Tell icinga2 to reload after stage config validation
            (defaults to ``true``).
        :param activate: Tell icinga2 to activate the stage if it validates
            (defaults to ``true``).
            If activate is set to false, reload must
            also be false.
        :param suppress_exception: If this parameter is set to ``True``, no
            exceptions are thrown.

        Example for a local configuration in the ``conf.d`` directory:

        .. code-block::

            raw_client.configuration.create_stage(
                package_name="example-cmdb",
                files={
                    "conf.d/test-host.conf": 'object Host "test-host" { address = "127.0.0.1", check_command = "hostalive" }'
                },
            )

        Example for a host configuration inside the ``satellite`` zone in the ``zones.d`` directory:

        .. code-block::

            raw_client.configuration.create_stage(
                package_name="example-cmdb",
                files={
                    "zones.d/satellite/host2.conf": 'object Host "satellite-host" { address = "192.168.1.100", check_command = "hostalive"',
                },
            )

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#create-a-stage-upload-configuration <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#create-a-stage-upload-configuration>`__
        """
        return self._request(
            "POST",
            f"stages/{_normalize_name(package_name)}",
            assemble_payload(files=files, reload=reload, activate=activate),
            suppress_exception=suppress_exception,
        )

    def list_packages(
        self,
        suppress_exception: Optional[bool] = None,
    ) -> Any:
        """
        List packages and their stages.

        :param suppress_exception: If this parameter is set to ``True``, no
            exceptions are thrown.

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#list-configuration-packages-and-their-stages <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#list-configuration-packages-and-their-stages>`__
        """
        return self._request(
            "GET",
            "packages",
            suppress_exception=suppress_exception,
        )

    def list_stage_files(
        self,
        package_name: str,
        stage_name: str,
        suppress_exception: Optional[bool] = None,
    ) -> Any:
        """
        List packages and their stages.

        :param package_name: Package names with the ``_`` prefix are reserved for
            internal packages and must not be used. You can recognize
            ``_api``, ``_etc`` and ``_cluster`` when querying specific objects
            and packages.
        :param stage_name: The stage name, for example ``7e7861c8-8008-4e8d-9910-2a0bb26921bd``.
        :param suppress_exception: If this parameter is set to ``True``, no
            exceptions are thrown.

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#list-configuration-package-stage-files <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#list-configuration-package-stage-files>`__
        """
        return self._request(
            "GET",
            f"stages/{package_name}/{stage_name}",
            suppress_exception=suppress_exception,
        )

    def delete_package(
        self,
        name: str,
        suppress_exception: Optional[bool] = None,
    ) -> Any:
        """
        :param name: Package names with the ``_`` prefix are reserved for
            internal packages and must not be used. You can recognize
            ``_api``, ``_etc`` and ``_cluster`` when querying specific objects
            and packages.
        :param suppress_exception: If this parameter is set to ``True``, no
            exceptions are thrown.

        :see: `Icinga2 API documentation: 12-icinga2-api/#deleting-configuration-package <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#deleting-configuration-package>`__
        """
        return self._request(
            "DELETE",
            f"packages/{_normalize_name(name)}",
            suppress_exception=suppress_exception,
        )


EventStreamType = Literal[
    "CheckResult",  # Check results for hosts and services.
    "StateChange",  # Host/service state changes.
    "Notification",  # Notification events including notified users for hosts and services.
    "AcknowledgementSet",  # Acknowledgement set on hosts and services.
    "AcknowledgementCleared",  # Acknowledgement cleared on hosts and services.
    "CommentAdded",  # Comment added for hosts and services.
    "CommentRemoved",  # Comment removed for hosts and services.
    "DowntimeAdded",  # Downtime added for hosts and services.
    "DowntimeRemoved",  # Downtime removed for hosts and services.
    "DowntimeStarted",  # Downtime started for hosts and services.
    "DowntimeTriggered",  # Downtime triggered for hosts and services.
    "ObjectCreated",  # Object created for all Icinga 2 objects.
    "ObjectDeleted",  # Object deleted for all Icinga 2 objects.
    "ObjectModified",  # Object modified for all Icinga 2 objects.
]
"""
:see: `doc/12-icinga2-api/#event-stream-types <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#event-stream-types>`__
:see: `lib/icinga/apievents.hpp L16-L47 <https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/icinga/apievents.hpp#L16-L47>`__
:see: `lib/remote/eventshandler.cpp L22-L38 <https://github.com/Icinga/icinga2/blob/c0b047b1aab6de3c5e51fdeb63d3bf4236f7fa6d/lib/remote/eventshandler.cpp#L22-L38>`__"""


class EventsUrlEndpoint(RequestHandler):
    """
    Connects to the URL endpoint ``events`` of the Icinga2 API.
    """

    path_prefix = "events"

    def subscribe(
        self,
        types: Sequence[EventStreamType],
        queue: str,
        filter: Optional[str] = None,
        filter_vars: FilterVars = None,
    ) -> Generator[str | Any, Any, None]:
        """
        Subscribe to an event stream.

        Event streams can be used to receive check results, downtimes,
        comments, acknowledgements, etc. as a “live stream” from Icinga.

        You can for example forward these types into your own backend. Process
        the metrics and correlate them with notifications and state changes
        e.g. in Elasticsearch with the help of Icingabeat. Another use case
        are aligned events and creating/resolving tickets automatically in
        your ticket system.

        You can subscribe to event streams by sending a POST request to the
        URL endpoint /v1/events. The following parameters need to be
        specified (either as URL parameters or in a JSON-encoded message body):

        example 1:

        .. code-block:: python

            types = ["CheckResult"]
            queue = "monitor"
            filters = "event.check_result.exit_status==2"
            for event in subscribe(types, queue, filters):
                print event

        :param types: **Required.** Event type(s). Multiple types as
            URL parameters are supported.
        :param queue: **Required.** Unique queue name. Multiple HTTP clients
            can use the same queue as long as they use the same event types
            and filter.
        :param filter: Filter for specific event attributes using
            filter expressions.
        :param filter_vars: variables used in the filters expression

        :returns: the events

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#event-streams <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#event-streams>`__
        """
        stream = self._request(
            "POST",
            None,
            assemble_payload(
                types=types, queue=queue, filter=filter, filter_vars=filter_vars
            ),
            stream=True,
        )
        for event in self._get_message_from_stream(stream):
            yield event


def _normalize_name(name: str) -> str:
    """To be able to send names with spaces or special characters to the REST API."""
    return urllib.parse.quote(name, safe="")


class Attrs:
    """https://github.com/Icinga/icinga2/blob/master/lib/icinga/checkable.ti"""

    __name: str
    acknowledgement: int
    acknowledgement_expiry: int
    acknowledgement_last_change: int
    action_url: str
    active: bool
    check_attempt: int
    check_command: str
    check_interval: int
    check_period: str
    check_timeout: None
    command_endpoint: str
    display_name: str
    downtime_depth: int
    enable_active_checks: bool
    enable_event_handler: bool
    enable_flapping: bool
    enable_notifications: bool
    enable_passive_checks: bool
    enable_perfdata: bool
    event_command: str
    executions: None
    flapping: bool
    flapping_current: int
    flapping_ignore_states: None
    flapping_last_change: int
    flapping_threshold: int
    flapping_threshold_high: int
    flapping_threshold_low: int
    force_next_check: bool
    force_next_notification: bool
    groups: list[str]
    ha_mode: int
    handled: bool
    host_name: str
    icon_image: str
    icon_image_alt: str
    last_check: float


class Object:
    attrs: dict[str, Any]
    joins: dict[str, Any]


class Service(Object):
    """https://github.com/Icinga/icinga2/blob/master/lib/icinga/service.ti"""

    type = "Service"
    name: str
    meta: dict[str, Any]


class Host:
    name: str
    state: int
    last_check_result: CheckResult


class ObjectsUrlEndpoint(RequestHandler):
    """
    Connects to the URL endpoint ``objects`` of the Icinga2 API.

    Provides methods to manage configuration objects:

    - creating objects
    - querying objects
    - modifying objects
    - deleting objects
    """

    path_prefix = "objects"

    def list(
        self,
        object_type: ObjectTypeName,
        name: Optional[str] = None,
        attrs: Optional[Sequence[str]] = None,
        filters: Optional[str] = None,
        filter_vars: FilterVars = None,
        joins: Optional[Union[bool, Sequence[str]]] = None,
        suppress_exception: Optional[bool] = None,
    ) -> Any:
        """
        get object by type or name

        :param object_type: The type of the object, for example ``Service``,
            ``Host`` or ``User``.
        :param name: The full object name, for example ``example.localdomain``
            or ``example.localdomain!http``.
        :param attrs: only return these attributes
        :param filters: filters matched object(s)
        :param filter_vars: variables used in the filters expression
        :param joins: show joined object
        :param suppress_exception: If this parameter is set to ``True``, no exceptions are thrown.

        Get all hosts:

        .. code-block:: python

            raw_client.objects.list("Host")

        List the service ``ping4`` of host ``webserver01.domain!ping4``:

        .. code-block:: python

            raw_client.objects.list("Service", "webserver01.domain!ping4")

        Get all hosts but limit attributes to `address` and `state`:

        .. code-block:: python

            raw_client.objects.list("Host", attrs=["address", "state"])

        Get all hosts which have ``webserver`` in their host name:

        .. code-block:: python

            raw_client.objects.list("Host", filters='match("webserver*", host.name)')

        Get all services and the joined host name:

        .. code-block:: python

            raw_client.objects.list("Service", joins=["host.name"])

        Get all services and all supported joins:

        .. code-block:: python

            raw_client.objects.list("Service", joins=True)

        Get all services which names start with ``vHost`` and are assigned to hosts named ``webserver*`` using ``filter_vars``:

        .. code-block:: python

            hostname_pattern = "webserver*"
            service_pattern = "vHost*"
            raw_client.objects.list(
                "Service",
                filters="match(hpattern, host.name) && match(spattern, service.name)",
                filter_vars={"hpattern": hostname_pattern, "spattern": service_pattern},
            )

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#querying-objects <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#querying-objects>`__
        """

        url_path = pluralize_to_lower_object_type_name(object_type)
        if url_path == "dependencys":
            url_path = "dependencies"
        if name:
            url_path += f"/{_normalize_name(name)}"

        payload: Payload = {}
        if attrs:
            payload["attrs"] = attrs
        if filters:
            payload["filter"] = filters
        if filter_vars:
            payload["filter_vars"] = filter_vars
        if isinstance(joins, bool) and joins:
            payload["all_joins"] = "1"
        elif joins:
            payload["joins"] = joins

        result = self._request(
            "GET", url_path, payload, suppress_exception=suppress_exception
        )
        if "results" in result:
            return result["results"]
        return result

    def get(
        self,
        object_type: ObjectTypeName,
        name: str,
        attrs: Optional[Sequence[str]] = None,
        joins: Optional[Union[bool, Sequence[str]]] = None,
        suppress_exception: Optional[bool] = None,
    ) -> Any:
        """
        Get a single object (``Host``, ``Service``, ...).

        :param object_type: The type of the object, for example ``Service``,
            ``Host`` or ``User``.
        :param name: The full object name, for example ``example.localdomain``
            or ``example.localdomain!http``.
        :param attrs:  Get only the specified objects attributes.
        :param joins: Also get the joined object, e.g. for a `Service` the `Host` object.

        :param suppress_exception: If this parameter is set to ``True``, no exceptions are thrown.

        Get host ``webserver01.domain``:

        .. code-block:: python

            raw_client.objects.get("Host", "webserver01.domain")

        Get service ``ping4`` of host ``webserver01.domain``:

        .. code-block:: python

            raw_client.objects.get("Service", "webserver01.domain!ping4")

        Get host ``webserver01.domain`` but the attributes ``address`` and ``state``:

        .. code-block:: python

            raw_client.objects.get("Host", "webserver01.domain", attrs=["address", "state"])

        Get service ``ping4`` of host ``webserver01.domain`` and the host attributes:

        .. code-block:: python

            raw_client.objects.get("Service", "webserver01.domain!ping4", joins=True)

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#querying-objects <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#querying-objects>`__
        """
        result = self.list(
            object_type, name, attrs, joins=joins, suppress_exception=suppress_exception
        )
        if "error" not in result:
            return result[0]

    def create(
        self,
        object_type: ObjectTypeName,
        name: str,
        templates: Optional[Sequence[str]] = None,
        attrs: Optional[Payload] = None,
        suppress_exception: Optional[bool] = None,
    ) -> Any:
        """
        Create an object using ``templates`` and specify attributes (``attrs``).

        :param object_type: The type of the object, for example ``Service``,
            ``Host`` or ``User``.
        :param name: The full object name, for example ``example.localdomain``
            or ``example.localdomain!http``.
        :param templates: Import existing configuration templates for this
            object type. Note: These templates must either be statically
            configured or provided in config packages.
        :param attrs: Set specific object attributes for this object type.
        :param suppress_exception: If this parameter is set to ``True``, no exceptions are thrown.

        Create a host:

        .. code-block:: python

            raw_client.objects.create(
                "Host", "localhost", ["generic-host"], {"address": "127.0.0.1"}
            )

        Create a service for Host ``localhost``:

        .. code-block:: python

            raw_client.objects.create(
                "Service",
                "testhost3!dummy",
                {"check_command": "dummy"},
                ["generic-service"],
            )

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#creating-config-objects <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#creating-config-objects>`__
        """

        payload: Payload = {}
        if attrs:
            payload["attrs"] = attrs
        if templates:
            payload["templates"] = templates

        return self._request(
            "PUT",
            f"{pluralize_to_lower_object_type_name(object_type)}/{_normalize_name(name)}",
            payload,
            suppress_exception=suppress_exception,
        )

    def update(
        self,
        object_type: ObjectTypeName,
        name: str,
        attrs: dict[str, Any],
        suppress_exception: Optional[bool] = None,
    ) -> Any:
        """
        Update an object with the specified attributes.

        :param object_type: The type of the object, for example ``Service``,
            ``Host`` or ``User``.
        :param name: the name of the object
        :param attrs: object's attributes to change
        :param suppress_exception: If this parameter is set to ``True``, no exceptions are thrown.

        Change the ip address of a host:

        .. code-block:: python

            raw_client.objects.update("Host", "localhost", {"address": "127.0.1.1"})

        Update a service and change the check interval:

        .. code-block:: python

            raw_client.objects.update(
                "Service", "testhost3!dummy", {"check_interval": "10m"}
            )

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#modifying-objects <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#modifying-objects>`__
        """
        return self._request(
            "POST",
            f"{pluralize_to_lower_object_type_name(object_type)}/{name}",
            attrs,
            suppress_exception=suppress_exception,
        )

    def delete(
        self,
        object_type: ObjectTypeName,
        name: Optional[str] = None,
        filters: Optional[str] = None,
        filter_vars: FilterVars = None,
        cascade: bool = True,
        suppress_exception: Optional[bool] = None,
    ) -> Any:
        """Delete an object.

        :param object_type: The type of the object, for example ``Service``,
            ``Host`` or ``User``.
        :param name: The full object name, for example ``example.localdomain``
            or ``example.localdomain!http``.
        :param filters: filters matched object(s)
        :param filter_vars: variables used in the filters expression
        :param cascade: Delete objects depending on the deleted objects (e.g. services on a host).
        :param suppress_exception: If this parameter is set to ``True``, no exceptions are thrown.

        Delete the ``localhost``:

        .. code-block:: python

            raw_client.objects.delete("Host", "localhost")

        Delete all services matching ``vhost*``:

        .. code-block:: python

            raw_client.objects.delete("Service", filters='match("vhost*", service.name)')

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#deleting-objects <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#deleting-objects>`__
        """

        object_type_url_path = pluralize_to_lower_object_type_name(object_type)

        payload: Payload = {}
        if filters:
            payload["filter"] = filters
        if filter_vars:
            payload["filter_vars"] = filter_vars
        if cascade:
            payload["cascade"] = 1

        url = object_type_url_path
        if name:
            url += f"/{_normalize_name(name)}"

        return self._request(
            "DELETE", url, payload, suppress_exception=suppress_exception
        )


StatusType = Literal[
    "ApiListener",
    "CIB",
    "CheckerComponent",
    "ElasticsearchWriter",
    "FileLogger",
    "GelfWriter",
    "GraphiteWriter",
    "IcingaApplication",
    "IdoMysqlConnection",
    "IdoPgsqlConnection",
    "Influxdb2Writer",
    "InfluxdbWriter",
    "JournaldLogger",
    "NotificationComponent",
    "OpenTsdbWriter",
    "PerfdataWriter",
    "SyslogLogger",
]


class StatusUrlEndpoint(RequestHandler):
    """
    Connects to the URL endpoint ``status`` of the Icinga2 API.

    :see: `lib/remote/statushandler.cpp <https://github.com/Icinga/icinga2/blob/master/lib/remote/statushandler.cpp>`__:
    """

    path_prefix = "status"

    def list(self, status_type: Optional[StatusType | str] = None) -> Any:
        """
        Retrieve status information and statistics for Icinga 2.

        List complete status::

        .. code-block:: python

            client.status.list()

        List the status of the core application:

        .. code-block:: python

            client.status.list("IcingaApplication")

        :param status_type: Limit the output by specifying a status type, e.g. ``IcingaApplication``.

        :returns: status information

        :see: `doc/12-icinga2-api/#status-and-statistics <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#status-and-statistics>`__
        """

        url: str = ""
        if status_type:
            url = status_type

        return self._request("GET", url)


@dataclass
class PerfdataValue:
    """
    `lib/base/perfdatavalue.ti L8-L18 <https://github.com/Icinga/icinga2/blob/4c6b93d61775ff98fc671b05ad4de2b62945ba6a/lib/base/perfdatavalue.ti#L8-L18>`_
    `lib/base/perfdatavalue.hpp L17-L36 <https://github.com/Icinga/icinga2/blob/4c6b93d61775ff98fc671b05ad4de2b62945ba6a/lib/base/perfdatavalue.hpp#L17-L36>`__
    """

    label: str
    value: float
    counter: bool
    unit: str
    crit: Optional[Value] = None
    warn: Optional[Value] = None
    min: Optional[Value] = None
    max: Optional[Value] = None


@dataclass
class StatusMessage:
    """
    :see: `lib/remote/statushandler.cpp L53-L57 <https://github.com/Icinga/icinga2/blob/4c6b93d61775ff98fc671b05ad4de2b62945ba6a/lib/remote/statushandler.cpp#L53-L57>`_
    """

    name: str

    status: dict[str, Any]

    perfdata: Optional[Sequence[PerfdataValue]]


class TemplatesUrlEndpoint(RequestHandler):
    """
    Connects to the URL endpoint ``templates`` of the Icinga2 API.
    """

    path_prefix = "templates"

    def list(self, object_type: ObjectTypeName, filter: Optional[str] = None) -> Any:
        """Request information about configuration templates.

        :param object_type: The type of the object, for example ``Service``,
            ``Host`` or ``User``.
        :param filter: The template object can be accessed in the filter using the
            ``tmpl`` variable. In the example ``"match(\"g*\", tmpl.name)"``
            the match function is used to check a wildcard string pattern against
            ``tmpl.name``.

        :see: `Icinga2 API documentation: doc/12-icinga2-api/#querying-templates <https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#querying-templates>`__
        """
        payload: Payload = {}
        if filter:
            payload["filter"] = filter
        return self._request(
            "GET",
            pluralize_to_lower_object_type_name(object_type),
            payload,
        )


class RawClient:
    """
    This raw client is a thin wrapper around the Icinga2 REST API.

    You can use the client with either username/password combination or using certificates.

    Example using username and password:

    .. code-block:: python

        from pretiac.client import Client

        client = Client("localhost", 5665, "username", "password")

    Example using certificates:

    .. code-block:: python

        client = Client(
            "localhost",
            5665,
            certificate="/etc/ssl/certs/myhostname.crt",
            key="/etc/ssl/keys/myhostname.key",
        )

    If your public and private are in the same file, just use the `certificate` parameter.

    To verify the server certificate specify a ca file as `ca_file` parameter.

    Example:

    .. code-block:: python

        from pretiac.client import Client

        client = Client(
            "https://icinga2:5665",
            certificate="/etc/ssl/certs/myhostname.crt",
            key="/etc/ssl/keys/myhostname.key",
            ca_file="/etc/ssl/certs/my_ca.crt",
        )

    """

    config: Config

    url: str

    version: str

    actions: ActionsUrlEndpoint
    """Connects to the URL endpoint ``actions`` of the Icinga2 API."""

    configuration: ConfigurationUrlEndpoint
    """Connects to the URL endpoint ``config`` of the Icinga2 API."""

    events: EventsUrlEndpoint
    """Connects to the URL endpoint ``events`` of the Icinga2 API."""

    objects: ObjectsUrlEndpoint
    """Connects to the URL endpoint ``objects`` of the Icinga2 API."""

    status: StatusUrlEndpoint
    """Connects to the URL endpoint ``status`` of the Icinga2 API."""

    templates: TemplatesUrlEndpoint
    """Connects to the URL endpoint ``templates`` of the Icinga2 API."""

    def __init__(self, config: Config) -> None:
        """
        initialize object

        :param suppress_exception: If this parameter is set to ``True``, no exceptions are thrown.
        """
        self.config = config

        self.url = (
            f"https://{self.config.api_endpoint_host}:{self.config.api_endpoint_port}"
        )

        self.version = get_version("pretiac")

        self.actions = ActionsUrlEndpoint(self)
        self.configuration = ConfigurationUrlEndpoint(self)
        self.events = EventsUrlEndpoint(self)
        self.objects = ObjectsUrlEndpoint(self)
        self.status = StatusUrlEndpoint(self)
        self.templates = TemplatesUrlEndpoint(self)



import os
from pathlib import Path
from typing import Literal, Optional, Sequence, Union

import yaml
from pydantic import TypeAdapter
from pydantic.dataclasses import dataclass

from pretiac.exceptions import PretiacException
from pretiac.object_types import Payload


@dataclass(config={"extra": "forbid"})
class ObjectConfig:
    """
    Bundles all configuration required to create an object.
    """

    templates: Optional[Sequence[str]] = None
    """
    Import existing configuration templates for this
    object type. Note: These templates must either be statically
    configured or provided in config packages.
    """

    attrs: Optional["Payload"] = None
    """Set specific object attributes for this object type."""


@dataclass(config={"extra": "forbid"})
class Config:
    """
    :see: `pretiac (JS) <https://github.com/Josef-Friedrich/PREtty-Typed-Icinga2-Api-Client_js/blob/722c6308d79f603a9ad7678609cd907b932c64ab/src/client.ts#L7-L15>`__
    """

    config_file: Optional[Path] = None
    """The path of the loaded configuration file."""

    api_endpoint_host: Optional[str] = None
    """
    The domain or the IP address of the API endpoint, e. g. ``icinga.example.com``,
    ``localhost`` or ``127.0.0.1``.
    """

    api_endpoint_port: Optional[int] = None
    """The TCP port of the API endpoint, for example ``5665``.

    :see: `Icinca Object Types (apilistener) <https://icinga.com/docs/icinga-2/latest/doc/09-object-types/#apilistener>`__
    """

    http_basic_username: Optional[str] = None
    """
    The name of the API user used in the HTTP basic authentification, e. g. ``apiuser``.

    .. code-block ::

        object ApiUser "apiuser" {
            ...
        }

    :see: `Icinca Object Types (apiuser) <https://icinga.com/docs/icinga-2/latest/doc/09-object-types/#apiuser>`__
    """

    http_basic_password: Optional[str] = None
    """
    The password of the API user used in the HTTP basic authentification, e. g. ``password``.

    .. code-block ::

        object ApiUser "apiuser" {
            password = "password"
        }

    :see: `Icinca Object Types <https://icinga.com/docs/icinga-2/latest/doc/09-object-types/#apiuser>`__
    """

    client_private_key: Optional[str] = None
    """
    The file path of the client’s **private RSA key**, for example
    ``/etc/pretiac/api-client.key.pem``.

    The RSA private key is created with this command:

    .. code-block ::

        icinga2 pki new-cert \\
            --cn api-client \\
            --key api-client.key.pem \\
            --csr api-client.csr.pem
    """

    client_certificate: Optional[str] = None
    """
    The file path of the client **certificate**.

    The certificate is created with this command:

    .. code-block ::

        icinga2 pki sign-csr \\
            --csr api-client.csr.pem \\
            --cert api-client.cert.pem
    """

    ca_certificate: Optional[str] = None
    """
    The file path of the Icinga **CA (Certification Authority)**.

    The CA certificate is located at ``/var/lib/icinga2/certs/ca.crt``. This
    command copies the certificate to the local host.

    .. code-block ::

        scp icinga-master:/var/lib/icinga2/certs/ca.crt .
    """

    suppress_exception: Optional[bool] = None
    """
    If set to ``True``, no exceptions are thrown.
    """

    new_host_defaults: Optional[ObjectConfig] = None
    """If a new host needs to be created, use this defaults."""

    new_service_defaults: Optional[ObjectConfig] = None
    """If a new service needs to be created, use this defaults."""

    def check(self) -> None:
        """Check if all required values are set."""
        if self.api_endpoint_host is None:
            raise PretiacException("Specify an API endpoint host (api_endpoint_host)!")

        if (self.http_basic_username or self.http_basic_password) and (
            self.client_private_key or self.client_certificate or self.ca_certificate
        ):
            raise PretiacException(
                "Specify HTTP basic OR certificate authentification. Not both!"
            )


def load_config_file(config_file: Optional[Union[str, Path]] = None) -> Config:
    """
    Load the configuration file in YAML format.

    The file path of the loaded configuration file is determined in this order:

    1. The parameter ``config_file`` of this function.
    2. The file path in the environment variable ``PRETIAC_CONFIG_FILE``.
    3. The configuration file in the home folder ``~/.pretiac.yml``.
    4. The configuration file at ``/etc/pretiac/config.yml``.

    :param config_file: The path of the configuration file to load.
    """
    config_files: list[Path] = []
    if config_file:
        if isinstance(config_file, str):
            config_files.append(Path(config_file))
        else:
            config_files.append(config_file)
    if "PRETIAC_CONFIG_FILE" in os.environ:
        config_files.append(Path(os.environ["PRETIAC_CONFIG_FILE"]))
    config_files.append(Path.cwd() / ".pretiac.yml")
    config_files.append(Path("/etc/pretiac/config.yml"))

    for path in config_files:
        if path.exists():
            config_file = path
            break

    adapter = TypeAdapter(Config)

    if not config_file:
        return adapter.validate_python({})

    with open(config_file, "r") as file:
        config_raw = yaml.safe_load(file)
        config_raw["config_file"] = str(config_file)
    return adapter.validate_python(config_raw)


def load_config(
    config: Optional[Config] = None,
    config_file: Optional[Union[str, Path, Literal[False]]] = None,
    api_endpoint_host: Optional[str] = None,
    api_endpoint_port: Optional[int] = None,
    http_basic_username: Optional[str] = None,
    http_basic_password: Optional[str] = None,
    client_private_key: Optional[str] = None,
    client_certificate: Optional[str] = None,
    ca_certificate: Optional[str] = None,
    suppress_exception: Optional[bool] = None,
    new_host_defaults: Optional[ObjectConfig] = None,
    new_service_defaults: Optional[ObjectConfig] = None,
) -> Config:
    """
    :param config: A configuration object that has already been populated.
    :param config_file: The path of the configuration file to load.
        If this value is set to false no configuration file will be loaded.
    :param api_endpoint_host: The domain or the IP address of the API
        endpoint, e. g. ``icinga.example.com``, ``localhost`` or ``127.0.0.1``.
    :param api_endpoint_port: The TCP port of the API endpoint, for example
        ``5665``.
    :param http_basic_username: The name of the API user used in the HTTP basic
        authentification, e. g. ``apiuser``.
    :param http_basic_password: The password of the API user used in the HTTP
        basic authentification, e. g. ``password``.
    :param client_private_key: The file path of the client’s **private RSA
        key**, for example ``/etc/pretiac/api-client.key.pem``.
    :param client_certificate: The file path of the client’s **certificate**,
        for example ``/etc/pretiac/api-client.cert.pem``.
    :param ca_certificate: The file path of the Icinga **CA (Certification
        Authority)**, for example ``/var/lib/icinga2/certs/ca.crt``.
    :param suppress_exception: If set to ``True``, no exceptions are thrown.
    :param new_host_defaults: If a new host needs to be created, use this
        defaults.
    :param new_service_defaults: If a new service needs to be created, use
        this defaults.
    """
    if config is not None and config_file is not None:
        raise PretiacException("Specify config OR config_file. Not both!")

    c: Optional[Config] = None
    if config:
        c = config
    elif config_file is False:
        c = Config()
    else:
        c = load_config_file(config_file)

    if api_endpoint_host is not None:
        c.api_endpoint_host = api_endpoint_host

    if api_endpoint_port is not None:
        c.api_endpoint_port = api_endpoint_port

    if c.api_endpoint_port is None:
        c.api_endpoint_port = 5665

    if http_basic_username is not None:
        c.http_basic_username = http_basic_username

    if http_basic_password is not None:
        c.http_basic_password = http_basic_password

    if client_private_key is not None:
        c.client_private_key = client_private_key

    if client_certificate is not None:
        c.client_certificate = client_certificate

    if ca_certificate is not None:
        c.ca_certificate = ca_certificate

    if suppress_exception is not None:
        c.suppress_exception = suppress_exception

    if new_host_defaults is not None:
        c.new_host_defaults = new_host_defaults

    if new_service_defaults is not None:
        c.new_service_defaults = new_service_defaults

    c.check()

    return c



"""
Execute checks using subprocess and send it via the API to the monitoring server.
"""

import shlex
import subprocess
import time
from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional, Union

import yaml
from pydantic import TypeAdapter

from pretiac import set_default_client
from pretiac.client import CheckError, CheckResponse
from pretiac.log import logger
from pretiac.object_types import ServiceState, get_service_state


class CheckExecution:
    check_command: Sequence[str]

    execution_start: float

    execution_end: float

    exit_status: ServiceState

    plugin_output: str

    performance_data: Optional[str] = None

    def __init__(self, check_command: Union[Sequence[str], str]) -> None:
        if isinstance(check_command, str):
            check_command = shlex.split(check_command)
        self.check_command = check_command
        try:
            self.execution_start = time.time()
            self.execution_end = self.execution_start
            process = subprocess.run(
                self.check_command, capture_output=True, encoding="utf-8"
            )
            self.execution_end = time.time()
            self.exit_status = get_service_state(process.returncode)
            output = process.stdout.strip()
            segments = output.split("|")
            self.plugin_output = segments[0].strip()
            if len(segments) > 1:
                self.performance_data = segments[1].strip()

            logger.debug(
                "CheckExecution: check_command: %s", " ".join(self.check_command)
            )
        except FileNotFoundError:
            self.exit_status = ServiceState.CRITICAL
            self.plugin_output = f"Plugin not found: {self.check_command[0]}"
        except Exception as e:
            self.exit_status = ServiceState.CRITICAL
            self.plugin_output = f"{e.__class__.__name__}: {e.args}"


@dataclass
class ServiceCheck:
    """
    .. code-block:: yaml

        - service: procs_zombie
          display_name: Zombie processes
          check_command: check_procs --warning=5 --critical=10 -s Z

    """

    service: str
    """The name of the service."""

    check_command: str

    host: Optional[str] = None
    """The name of the host."""

    display_name: Optional[str] = None
    """A short description of the service, if it needs to be created."""

    def set_host(self, host: Optional[str]) -> None:
        if self.host is None:
            self.host = host

    def check(self) -> CheckResponse | CheckError:
        """Check and send the check result to the monitoring endpoint using the API."""
        check = CheckExecution(self.check_command)
        return set_default_client().send_service_check_result(
            service=self.service,
            host=self.host,
            exit_status=check.exit_status,
            execution_start=check.execution_start,
            execution_end=check.execution_end,
            check_command=check.check_command,
            plugin_output=check.plugin_output,
            performance_data=check.performance_data,
            display_name=self.display_name,
        )


@dataclass
class CheckCollection:
    """
    .. code-block:: yaml

        ---
        host: wrasp-passive

        checks:

        - service: memory
            check_command: check_linux_memory -f -w 2 -c 0

        - service: procs_zombie
            display_name: Zombie processes
            check_command: check_procs --warning=5 --critical=10 -s Z

    """

    checks: Sequence[ServiceCheck]
    host: Optional[str] = None

    def run_checks(self) -> None:
        for check in self.checks:
            check.set_host(self.host)
            check.check()


def _read_yaml(file_path: str | Path) -> Any:
    if isinstance(file_path, str):
        file_path = Path(file_path)
    if not file_path.exists():
        raise FileNotFoundError(f"{file_path} does not exist.")
    with open(file_path, "r") as file:
        data = yaml.safe_load(file)
    return data


def _read_check_collection(file_path: str | Path) -> CheckCollection:
    adapter = TypeAdapter(CheckCollection)
    return adapter.validate_python(_read_yaml(file_path))


def check(file_path: str | Path | None) -> None:
    logger.info("Read check collection file: %s", file_path)
    if file_path is None:
        file_path = "/etc/pretiac/checks.yml"
    collection: CheckCollection = _read_check_collection(file_path)
    collection.run_checks()
